/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _CBP_H
#define _CBP_H
#include "controlflow.h"
#include "tbsymbol.h"
#include "globalmachine.h"
#include "stringmanip.h"

#define _GLOBVAR 1
#define _LOCVAR 2
#define _LITTRUE 3
#define _LITFALSE 4
#define _STAR 5

#define _version_ "0.2"

#include "error.h"

#include <string.h>

extern FILE * yyin ;
	
int lineno ;

/* Various symbol tables */

//Some global tables: they will be shared among all threads:
T_tbsymbol * _lock_table ;
T_tbsymbol * _glob_var_table ;
T_tbsymbol * _label_table ;
T_tbsymbol * _thread_table ;
T_tbsymbol * _msg_table ;

//These tables are local to the threads
T_tbsymbol * _temp_table ;
T_tbsymbol * _loc_var_table ;


//some global variable used to store the initial local states of the local machines
//char *current_thread;
//PTR_LM_list_initial_states list_of_initial_states;


/* This struct is used to store all the message that need to be inserted in the 'Actions' section of the GM file*/
typedef struct s_msgList {
	char * message ;
	struct s_msgList * next ;
	} msgList ;
	
typedef msgList * PTR_msgList ;

extern PTR_msgList mList ;

/* An insertion function related to this structure */
void msgList_insert(char * msg, PTR_msgList * l) ;

//This datastructure will be the type of the info field in the thread symbol table. It can retain the control flow graph and
//the local variable symbol table of the thread

typedef struct s_ThreadInfo{
	T_tbsymbol * vartable ;
	CF_graph * cfg ;
	PTR_LM_initial_states initStates ;
	PTR_local_machine locMach ;
/*	struct s_ThreadInfo * next ;*/
	} ThreadInfo ;

// The label counter we use when we need to forge additional labels (for the instructions that don't have one)
int lblcnt ;
	
// one control flow graph
CF_graph * cfg ;

typedef struct {
	CF_PTR_node begin ;
	CF_PTR_arc out ;
	} sub_graph ;

/*typedef struct ElseIfRet {
 *	CF_PTR_node begin ;
 *	CF_PTR_node dummy ; // The dummy node at the end of a if.. elseif block 
 *	} ElseIfRet ;
 */

/* A list of arcs (each arc is a guarded assignation) */
typedef struct GuardedAssignList {
	CF_PTR_arc arc ; /* The arc representing the guarded assignation */
	struct GuardedAssignList * next ;
} GuardedAssignList ;


//Return type of expression
typedef struct ExpressionRet {
	int type ; // type of the variable : _GLOBVAR or _LOCVAR or _LITTRUE or _LITFALSE or _STAR
	int var ; // the entry in the symbol table for the variable
	int not ; // says wether there is a ! before the variable/the litteral or not 0= no ! and 1= there is a !
	} ExpressionRet ;
	


//Return type of ExpressionList
typedef struct ExpListRet {
	struct ExpressionRet * exp ; //expression of the element of the list
	struct ExpListRet * next; //next element of the list	
} ExpListRet ;

//Return type of Variables
typedef struct s_VarRet {
	int varId ;
	struct s_VarRet * next ; // It is simply a list of variable id's
	} VarRet  ;
	
//Return type of assign
typedef struct s_AssignRet {
	int varId ;
	int varType;
	ExpressionRet * exp ;
	struct s_AssignRet * next ;
	} AssignRet ;
	
//Return types of assign Middle:

//1 - The variable list
typedef struct s_AssignVarList {
	int varId ;
	int varType ;
	struct s_AssignVarList * next ;
	} AssignVarList ;
	
// 2 - The expression list
typedef struct s_AssignExpList {
	ExpressionRet * exp ;
	struct s_AssignExpList * next ;
	} AssignExpList ;

// 3 - The type combining both of them
typedef struct s_AssignMiddleRet {
	AssignExpList * expList ;
	AssignVarList * varList ;
	} AssignMiddleRet ;
	
/*This structure is used to retain two types of informations in the info fields of the choice arcs */
typedef struct GuardedAssignInfo {
	AssignRet * assign ; /* The assignation to perform */
	ExpressionRet *  test ; /* The test to perform (if any) */
} GuardedAssignInfo ;

	
int yyerror(char * s);

sub_graph __if(ExpressionRet * expression, sub_graph * ifbranch, sub_graph * elsebranch) ;
sub_graph __while(ExpressionRet * expression, sub_graph _exp) ;
sub_graph __skip(void) ;
sub_graph __accept( char * msg) ;
sub_graph __wakeupall(char * msg) ;
sub_graph __wakeup(char * msg) ;
sub_graph __start(char * msg) ;
sub_graph __rdv( char * msg) ;
sub_graph __sleep( char *lock,char *msg) ;
sub_graph __unlock( char *lock) ;
sub_graph __lock(char *lock) ;
sub_graph __goto( char * label) ;
sub_graph __rdv_param(char * msg, ExpListRet * param) ;
sub_graph __start_param(char * msg, char * exp) ;
sub_graph __accept_param(char * msg, VarRet * vars) ;
sub_graph __assign(AssignRet * assign) ;
void subList(AssignRet ** List, AssignRet ** l1) ;
void rec_build_arcs(CF_PTR_node n, CF_PTR_node e, AssignRet * l, char * buf_a, char * buf_g) ;
sub_graph __guarded_assign(GuardedAssignList * list) ;
GuardedAssignList * ga_build_arc(ExpressionRet * exp, AssignRet * assign) ;
void list_assignation_literals_to_global(AssignRet ** assign, AssignRet ** lit) ;
#endif
