/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
%{
	#include "tbsymbol.h"
	#include "controlflow.h"
	#include "CBP.h"
	#define MAX_SYMBOL_SIZE 512
	/*ThreadInfo * setOfThread = NULL ;*/
	int yylex();

%}
%union{
	int val ;
	char * name ;
	sub_graph sg ;
	ExpressionRet * exp ;
	ExpListRet * explist ;
	VarRet * var ;
	ThreadInfo * info ;
	AssignRet * assign ;
	AssignMiddleRet * assignMid ;
/*	ElseIfRet * eif ;*/
	GuardedAssignList * ga ;
}

%token VARS
%token LOCKS
%token MESSAGES
%token THREADS
%token IF
%token ELSE
%token WHILE
%token SKIP
%token GOTO
%token LOCK
%token UNLOCK
%token SLEEP
%token WAKEUP
%token WAKEUPALL
%token START
%token RDV
%token ACCEPT
%token TRUE
%token FALSE
%token EQUAL
%token ELSEIF
%token CHOICE
%token <name> ident
%token <name> number

%type <sg> Command
%type <sg> LabeledCommands
%type <name> Label
%type <exp> Expression
%type <val> Litteral
%type <exp> ExpressionStar
%type <name> Parameters
%type <explist> ExpressionList
%type <var> VariableList
%type <info> ThreadBody
%type <assign> Assignation
%type <assignMid> AssignMiddle
%type <ga> GuardedAssignList 
%type <ga> GuardedAssign 
/*%type <eif> ElseIf*/
%%

Program	:	Declarations Threads			{}
	;

Declarations	:	VarDec LockDec MsgDec ThreadDec		{
				_glob_var_table=_temp_table ;
				_temp_table = NULL ;
				
				}
		;
		
VarDec	:	VARS ':' Variables ';'			{}
	;

Variables	:	 Variables ident		{
				if (! _temp_table)
					tbsymbol_init(&_temp_table,MAX_SYMBOL_SIZE) ;
				tbsymbol_insert(_temp_table, $2) ;
				
				}
		|					{}
		;
		
LockDec		:	LOCKS ':' Locks ';'		{}
		;
		
Locks		:	ident Locks			{tbsymbol_insert(_lock_table, $1) ;}
		|					{}
		;

MsgDec		:	MESSAGES ':' Messages ';'	{}
		;

Messages	:	ident Messages			{tbsymbol_insert(_msg_table, $1) ;}
		|
		;

ThreadDec	:	THREADS ':' ThreadList ';'	{}
		;
		
ThreadList	:	ident ThreadList		{
	   							tbsymbol_insert(_thread_table, $1) ;
								//creation of an entry to store the initial states.
								LM_insert_list_local_states(list_of_initial_states, $1);
							}
		|					{}
		;

Threads		:	ident '{' ThreadBody '}'		{
				T_PTR_tbsymbol_entry p ;
				if ((p=tbsymbol_select(_thread_table, $1))==NULL)
					err_sys("[%i] I found thread %s's body, but %s is never declared. Giving up !", lineno, $1, $1) ;
				tbsymbol_setinfo(p, $3, sizeof($3)) ;
				
				/* Let's insert the threadInfo in the setOfThread list 
				$3->next = setOfThread ;
				setOfThread = $3 ;*/
				
				/* Let's associate the initial states to this thread */
				$3->initStates = LM_get_local_states(*list_of_initial_states, $1);
			}
			
		|	ident '{' ThreadBody '}' Threads	{
				T_PTR_tbsymbol_entry p ;
				if ((p=tbsymbol_select(_thread_table, $1))==NULL)
					err_sys("[%i] I found thread %s's body, but %s is never declared. Giving up !", lineno, $1, $1) ;
				tbsymbol_setinfo(p, $3, sizeof($3)) ;
				
				/* Let's insert the threadInfo in the setOfThread list 
				$3->next = setOfThread ;
				setOfThread = $3 ;*/
				
				/* Let's associate the initial states to this thread */
				$3->initStates = LM_get_local_states(*list_of_initial_states, $1);
		
			}
		;

ThreadBody	:	VarDec LabeledCommands			{
				/*We have now parsed a complete thread. We need to retain its control graph and its
				local variables table for subsequent use*/
				
				ThreadInfo * ti =(ThreadInfo*)malloc(sizeof(ThreadInfo)) ;
				CF_PTR_node node;
				ti->vartable = _temp_table ;
				_temp_table = NULL ;

				//adding of an ending node
				ti->cfg = cfg ;
				node = CF_create_node(-1,NULL,_NORMAL_NODE);
				CF_set_extr($2.out, node);
				CF_insert_node(cfg, node);


				//And we associate ti to the thread by returning it
				$$ = ti ;

				//Let's start a new graph
				CF_create_graph(&cfg) ;
				
				
				
				}
		;

Label		:	ident				{$$ = $1 ;}
		|	number				{$$ = $1 ;}
		;

LabeledCommands	:	'[' Label ']' Command LabeledCommands	{
						CF_PTR_node p ;
						//let's look for the label in the label symbol table
						int i = tbsymbol_get_index(_label_table, $2) ;
						if (i == -1) // it might already be there as we might have done a forward goto
						{
							//printf("yyparse: inserting label %s at position %i\n", $2, i) ; 
							i = tbsymbol_insert_num(_label_table, $2 ); // insert in the table
							CF_set_label_node($4.begin, i) ; // set the label of the first node
							$$.begin = $4.begin ; // no change, the begining node is the first of $4
						}
						else
						{	
							// we have to replace the first node of Command by the existing node
							p = CF_get_node(cfg, i) ;
							//printf("yyparse: found node %s at position %i (address %x)\n", $2, i, p) ;
							
							CF_merge_nodes(p, $4.begin) ;
							$$.begin = p ; // the first node is now p
						}
						
						if (CF_get_extr($4.out) == NULL) // if there's already an extremity, we are in a goto
							CF_set_extr($4.out, $5.begin) ; // attach the set of labeld commands to the graph
						$$.out=$5.out ; 
						cfg->first = $$.begin ;} 
		|	'[' Label ']' Command			{
						CF_PTR_node p ;
						//let's look for the label in the label symbol table
						int i = tbsymbol_get_index(_label_table, $2) ;
						if (i == -1) // it might already be there as we might have done a forward goto
						{
							//printf("yyparse: inserting label %s at position %i\n ", $2, i) ; 
							i = tbsymbol_insert_num(_label_table, $2 ); // insert in the table
							CF_set_label_node($4.begin, i) ;
							$$ = $4 ; // we just take $4 as the value to return
						}
						else
						{
							// we have to replace the first node of Command by the existing node
							p = CF_get_node(cfg, i) ;
							//printf("yyparse: found node %s at position %i (address %x)\n", $2, i, p) ;
							
							CF_merge_nodes(p, $4.begin) ;
							$$.begin = p ; // p is now the begining node
							$$.out = $4.out ; // but the outgoing arc remains the same
						}
						
						cfg->first = $$.begin ;} 
		|	Command LabeledCommands	{
						// let's forge a new label
						int i ;
						char * buf=(char *)malloc(sizeof(char)*15) ;
						sprintf(buf, "_CBP2GM_%d_",lblcnt) ;
						++lblcnt ;
						
						i = tbsymbol_get_index(_label_table, buf) ;						
						if (i != -1) // it shouldn't be there yet !
						{
							printf ("?? %s ??", buf) ;
							yyerror("label used twice !\n") ;
						}
						i = tbsymbol_insert_num(_label_table, buf ); // insert in the table
						CF_set_label_node($1.begin, i) ; // set the label of the first node
						if (CF_get_extr($1.out) == NULL) // if there's already an extremity, we are in a goto
							CF_set_extr($1.out, $2.begin) ; 
						$$.out=$2.out ; 
						$$.begin=$1.begin ;
						cfg->first = $$.begin ;} 
		|	Command			{
						// let's forge a new label
						int i ;
						char * buf=(char *)malloc(sizeof(char)*15) ;
						sprintf(buf, "_CBP2GM_%d_",lblcnt) ;
						++lblcnt ;
						
						i = tbsymbol_get_index(_label_table, buf) ;						
						if (i != -1) // it shouldn't be there yet !
						{
							printf ("?? %s ??", buf) ;
							yyerror("label used twice !\n") ;
						}
						i = tbsymbol_insert_num(_label_table, buf ); // insert in the table
						CF_set_label_node($1.begin, i) ; // set the label of the first node
						$$=$1 ;
						cfg->first = $$.begin ;}
		;

Command		:	Assignation ';'				{$$ = __assign($1) ;}
		|	IF '(' ExpressionStar ')' '{' LabeledCommands '}' ELSE '{' LabeledCommands '}'	{$$ = __if($3,&$6,&$10);}
		|	IF '(' ExpressionStar ')' '{' LabeledCommands '}' 	{$$ = __if($3,&$6,NULL);}
		|	CHOICE '{' GuardedAssignList '}' {$$ = __guarded_assign($3) ;}
		|	WHILE '(' ExpressionStar ')' '{' LabeledCommands '}'		{$$ = __while($3,$6);}
		|	SKIP ';'				{$$ = __skip() ;}
		|	GOTO '(' Label ')' ';'			{$$ = __goto($3) ;}
		|	LOCK '(' ident ')' ';'			{$$ = __lock($3);}
		|	UNLOCK '(' ident ')' ';'		{$$ = __unlock($3);}
		|	SLEEP '(' ident ',' ident ')' ';'	{$$ = __sleep($3,$5);}
		|	WAKEUP '(' ident ')' ';'		{$$ = __wakeup($3);}
		|	WAKEUPALL '(' ident ')' ';'		{$$ = __wakeupall($3) ;}
		|	START '(' ident ')' ';'			{
								PTR_LM_initial_states local_machine;
								LM_PTR_state state;
								$$ = __start($3) ;
								/* The set of already computed local states for this thread */
								local_machine = LM_get_local_states(*list_of_initial_states,$3);
								state = LM_create_state(NULL,NULL,0);
								/*We add the valuation to this set */
								if (local_machine->first_state == NULL) {
									/*printf("Laurent : insertion etat initial bidon\n");*/
									LM_insert_local_state(state,local_machine);
								}
								else {
									free(state);
								}
								local_machine->nb_variable = 0;
								}
		|	START '(' ident ',' Parameters ')' ';'		{
						PTR_LM_initial_states local_machine;
						LM_PTR_state state;
						int nb_value = strlen($5); 
						int *value = (int *)malloc (nb_value * sizeof(int));
						int current_value;
						$$ = __start_param($3, $5) ;
						//insertion of the local state
						for (current_value = 0; current_value < nb_value; current_value ++) {
							if('T' == $5[current_value])
								value[current_value] = 1;
							else
								value[current_value] = 0;
							}
						/* The set of already computed local states for this thread */
						local_machine = LM_get_local_states(*list_of_initial_states,$3);
						state = LM_create_state(NULL,value,nb_value);
						/*We add the valuation to this set */
						if (LM_contain_state(state, local_machine) == 0) {
							LM_insert_local_state(state,local_machine);
						}
						else {
							free(value);
							free(state);
						}
						local_machine->nb_variable = nb_value ;
					} 
		|	RDV '(' ident ')' ';'			{$$ = __rdv($3) ;}
		|	RDV '(' ident ',' ExpressionList ')' ';'	{$$ = __rdv_param($3, $5) ;} 
		| 	ACCEPT '(' ident ')' ';'		{$$ = __accept($3) ;}
		|	ACCEPT '(' ident ',' VariableList ')' ';'	{$$ = __accept_param($3, $5) ;} 
		;
		
/*Else 		:	ELSE '{' LabeledCommands '}' { 
	 										$$ = __else($3);
											}										     }
		|	{ 	$$ = (ElseIfRet *)malloc(sizeof(ElseIfRet)) ;
				$$->begin = CF_create_node(-1, NULL, _IF_NODE) ;
				$$->dummy = CF_create_node(-1, NULL, _NORMAL_NODE) ;
				CF_insert_node(cfg, $$->begin) ;
				CF_insert_node(cfg, $$->dummy) ;
			}
		;*/

GuardedAssignList 	:	GuardedAssign GuardedAssignList	{
									if ($1 != NULL) 
									{ /* $1 == NULL if the conditionon the branch is false*/
										$1->next = $2 ;
										$$ = $1 ;
									} else { $$ = $2 ; }
								}
			|	GuardedAssign			{$$ = $1 ;}
			;
			
GuardedAssign		:	ExpressionStar ':' Assignation ';' {$$ = ga_build_arc($1, $3) ;}
			;

ExpressionList	:	Expression ',' ExpressionList	{
							$$ = (ExpListRet *)malloc(sizeof(ExpListRet)) ;
	       						$$->exp =  $1;
							$$->next = $3;
	       						}
		|	Expression			{
							$$ = (ExpListRet *)malloc(sizeof(ExpListRet)) ;
	       						$$->exp =  $1;
							$$->next = NULL;
							}
		;
VariableList	:	ident ',' VariableList		{
				
				int i ;
				VarRet * elem = (VarRet*)malloc(sizeof(VarRet)) ;
				// Let's find this id in the variable table
				if ((i=tbsymbol_get_index(_temp_table, $1)) == -1)
				{	
					err_sys("[%i] Variable %s in accepts satement has never been declared as local...", lineno, $1) ;
				}
					
				// The variable has been found: we can insert it in the head of the list
				
				elem->next = $3 ;
				elem->varId = i ;

				$$=elem ;

				}
		|	ident				{ 
		
				int i ;
				VarRet * elem = (VarRet*)malloc(sizeof(VarRet)) ;
				// Let's find this id in the variable table
				if ((i=tbsymbol_get_index(_temp_table, $1)) == -1)
				{	
					err_sys("[%i] Variable %s in accepts satement has never been declared as local...", lineno, $1) ;
				}	
				//Let's create the tail of the list
				elem->next = NULL ;
				elem->varId = i ;
				$$ = elem ;
				}

		;

Assignation	:	ident AssignMiddle Expression 		{


			/*We receive two lists (one for the expression, one for the variables) through
			the AssignMiddle rule.
			1/ We have to add the first variable and the last expression to these lists
			2/ We have to merge them in a single list that is suitable with
			   the return type of Assignation rule*/
			   
			AssignVarList * v = (AssignVarList *)malloc(sizeof(AssignVarList)) ;
			AssignVarList * p ;
			AssignExpList * e = (AssignExpList *)malloc(sizeof(AssignExpList)) ;
			AssignExpList * q ;
			AssignRet * r ;
			int i ;
		
			/* Let's look for this id in the variable table*/
        	        if (_temp_table == NULL || (i=tbsymbol_get_index(_temp_table, $1)) == -1) {
				if ((i=tbsymbol_get_index(_glob_var_table, $1)) == -1)
                      		{
	                              	err_sys("[%i] Variable %s in assignation satement has never been declared...", lineno, $2) ;
        	                }
				else
					v->varType = _GLOBVAR;
			}	
			else
				v->varType = _LOCVAR;
			v->next = NULL ;
			v->varId = i ;
			
			/*As said before, we make a tail insertion*/		
			if ($2->varList == NULL)
				$2->varList = v ;
			else
			{
				p = $2->varList ;
				while (p->next) /* to the tail of the list...*/
					p = p->next ;
				p->next = v ;
			}
		
			/* Now, we insert the expression in the head of the exp list*/
			e->exp = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ;
			e->exp->type = $3->type ;
			e->exp->var = $3->var ;
			e->exp->not = $3->not ;
			
			e->next = $2->expList ;
			$2->expList = e ;
			
			
			/*Now, let's go through these two lists and create a new one*/
			
			p = $2->varList ;
			q = $2->expList ;
			
			/*head insertion*/
			$$ = (AssignRet *)malloc(sizeof(AssignRet)) ;
			$$->varId = p->varId ;
			$$->varType = p->varType ;
			$$->exp = q->exp ;
			
			p = p->next ;
			q = q->next ; 
			r = $$ ;
			
			while (p && q) /*obviously, they should have the same length !!*/
			{
				r->next = (AssignRet *)malloc(sizeof(AssignRet)) ;
				r = r->next ;
				r->varId = p->varId ;
				r->varType = p->varType ;
				r->exp = q->exp ;
				p = p->next ;
				q = q->next ;

			}
			
			r->next = NULL ;
			
			/*And we are done */
			
			
		}
		;
		
AssignMiddle	:	EQUAL		{
	     							
						/*We create the structure with both lists empty*/
	     					$$ = (AssignMiddleRet *)malloc(sizeof(AssignMiddleRet)) ;
						$$->expList = NULL ;
						$$->varList = NULL ;
					}
								
		|	',' ident AssignMiddle Expression ','	{
		
		/*We have to add the variable id to the tail of the varList,
		and the expression to the head of the expList */
		
		AssignVarList * v = (AssignVarList *)malloc(sizeof(AssignVarList)) ;
		AssignVarList * p ;
		AssignExpList * e = (AssignExpList *)malloc(sizeof(AssignExpList)) ;
		int i ;
		
		/* Let's look for this id in the variable table*/
		if (! _temp_table) tbsymbol_init(&_temp_table,MAX_SYMBOL_SIZE) ;
                if ((i=tbsymbol_get_index(_temp_table, $2)) == -1) {
			if ((i=tbsymbol_get_index(_glob_var_table, $2)) == -1)
                      	{
                              	err_sys("[%i] Variable %s in assignation satement has never been declared...", lineno, $2) ;
                        }
			else
				v->varType = _GLOBVAR;
		}
		else
			v->varType = _LOCVAR;
			v->next = NULL ;
		v->varId = i ;
		
		/*As said before, we make a tail insertion*/		
		if ($3->varList == NULL)
			$3->varList = v ;
		else
		{
			p = $3->varList ;
			while (p->next) /* to the tail of the list...*/
				p = p->next ;
			p->next = v ;
		}
		
		/* Now, we insert the expression in the head of the exp list*/
		e->exp = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ;
		e->exp->type = $4->type ;
		e->exp->var = $4->var ;
		e->exp->not = $4->not ;
		e->next = $3->expList ;
		$3->expList = e ;
		
		$$ = $3 ;
		}	
		;

Parameters	:	Litteral ',' Parameters	{
					$$ = (char *)malloc (sizeof($3)+2*sizeof(char)) ;
					if ($1 == _LITTRUE)
						strcpy($$, "T") ;
					else
						strcpy($$, "F") ;
					strcat($$, $3) ;}
		|	Litteral		{
						$$ = (char *)malloc(sizeof(char));
						if ($1 == _LITTRUE)
							strcpy($$, "T");
						else
							strcpy($$, "F") ;}
		;
		
ExpressionStar	:	'*'		{$$ = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ;
					$$->type = _STAR ;}

		|	Expression	{$$=$1 ;}
		;
		
Expression	:	ident		{ 	int i ;
						$$ = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ;
						$$->not = 0 ;
						// is it a local variable ?
						if (_temp_table == NULL || (i=tbsymbol_get_index(_temp_table, $1)) == -1)
						{
							// no, should be global
							if ((i=tbsymbol_get_index(_glob_var_table, $1)) == -1)
							{
								//oops, it's not declared
								err_sys("[%i] Undeclared variable %s", lineno, $1) ;
							}
							else
							{
								$$->type = _GLOBVAR ; // ok, it's global
							}
						}
						else
						{
							$$->type = _LOCVAR ; // local variable
						}
						// i now contains the index
						$$->var = i ;
					}
		|	'!' ident	{ 	int i ;
						$$ = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ;
						$$->not = 1 ;
						// is it a local variable ?
						if (_temp_table == NULL || (i=tbsymbol_get_index(_temp_table, $2)) == -1)
						{
							// no, should be global
							if ((i=tbsymbol_get_index(_glob_var_table, $2)) == -1)
							{
								//oops, it's not declared
								err_sys("[%i] Undeclared variable %s", lineno, $2) ;
							}
							else
							{
								$$->type = _GLOBVAR ; // ok, it's global
							}
						}
						else
						{
							$$->type = _LOCVAR ; // local variable
						}
						// i now contains the index
						$$->var = i ;

						
					}
		|	Litteral	{ $$ = (ExpressionRet *)malloc(sizeof(ExpressionRet)) ; $$->type = $1 ;}
		;

Litteral	:	TRUE		{ $$ = _LITTRUE ;}
		|	FALSE		{ $$ = _LITFALSE ;}
		;

%%

int yyerror(char * s)
{
	
	printf("Erreur !: ligne %d: %s\n\n",lineno, s) ;
	exit(0);
}

int main(int argc, char **argv)
{
	PTR_global_machine G ;
	char * outFileName = (char *)malloc(256*sizeof(char)) ;
	FILE * outFile ;
	lblcnt = 0 ;
		
	printf("\n*** This is CBP2GM compiler v %s (compiled %s %s) ***\n", _version_,  __DATE__, __TIME__) ;
	
	if(argc != 2)
	{
		err_sys("I'm not happy with these arguments you gave me. Please tell me a file name to open !!\n") ;
	}
	
	if (!(yyin = fopen(argv[1], "r")))
	{
		err_sys("Couldn't open file %s. Giving up !", argv[1]) ;
	}
	
	if (strcasecmp((argv[1]+strlen(argv[1])-4), ".cbp") == 0) /* Input file has extension .cbp */
	{
		strcpy(outFileName, argv[1]) ;
		outFileName[strlen(outFileName)-3]='g' ;
		outFileName[strlen(outFileName)-2]='m' ;
		outFileName[strlen(outFileName)-1]=0 ;
	}
	
	printf("\nInitializing internal structures...") ;
	/*Initialisation of the symbol tables*/
	_temp_table = NULL ;
	tbsymbol_init(&_lock_table,MAX_SYMBOL_SIZE) ;
	tbsymbol_init(&_msg_table,MAX_SYMBOL_SIZE) ;
	tbsymbol_init(&_thread_table,MAX_SYMBOL_SIZE) ;
	tbsymbol_init(&_label_table,MAX_SYMBOL_SIZE) ;

	//creation of the list of initial states
	create_list_local_states(&list_of_initial_states);

	CF_create_graph (&cfg) ;
	
	printf(" done !\n\n") ;
	printf("Parsing has started\n") ;

	lineno = 1 ;
	yyparse() ;
	printf("Parsing is finished !\n\n") ;

	printf("Generation of local machines has started...\n") ;
	G = GM_generate() ;
	printf("Generation of local machines is finished !\n\n") ;
	
	printf("Generating %s...\n", outFileName) ;
	
	if(!(outFile = fopen(outFileName, "w+")))
	{
		err_sys("Couldn't open file %s. Giving up !", outFileName) ;
	}
	
	
	GM_write_to_file(G, outFile, argv[1]) ;
	printf("Finished !\n") ;

	fclose(yyin) ;
	printf("Exiting \n\n") ;
	return 0;
	
}
