/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/***********************************************

Control graph data structure

************************************************/
#include "controlflow.h"
#include "malloc.h"

void CF_create_graph (CF_PTR_graph * graph) 
{
	*graph = (CF_PTR_graph)malloc(sizeof(CF_graph));
	(*graph)->head = NULL;
}

void CF_insert_node(CF_PTR_graph graph, CF_PTR_node node) 
{
	CF_PTR_node node_graph;
	node->next = NULL;
	node_graph = graph->head;
	if (!graph->head)
	{
		graph->head = node ;
	}
	else
	{	while (node_graph->next != NULL) {
			node_graph = node_graph->next;
		}
		node_graph->next = node;
	}
}

CF_PTR_node CF_create_node(int label, void * info, int mark) 
{
	CF_PTR_node node = (CF_PTR_node)malloc(sizeof(CF_node));
	node->label = label;
	node->f_arc = NULL;
	node->info = info;
	node->next = NULL;
	node->mark = mark ;
	node->aux = 0 ;
	return node;
}

CF_PTR_node CF_get_node(CF_PTR_graph graph, int label)
{
	CF_PTR_node node_graph;
	if (graph->head == NULL)
		return NULL;
	else {
		node_graph = graph->head;
        	while ((node_graph->label != label) && (node_graph->next != NULL)) {
                	node_graph = node_graph->next;
	        }
		if (node_graph->label == label)
			return node_graph;
		else
			return NULL;
	}
		
}
 
void CF_set_info(CF_PTR_node node, void * info, int mark) 
{
	node->mark = mark;
	if (node->info != NULL)
		free(node->info);
	node->info = info;
}

void CF_set_label_node(CF_PTR_node node, int label) 
{
	node->label = label;
}

/////////////////////////////////////////////////////////

void * CF_get_info(CF_PTR_node node)
{
	return node->info ;
}

void CF_insert_arc(CF_PTR_graph graph, CF_PTR_node orig, CF_PTR_arc arc) 
{
	CF_PTR_node p = graph->head ;
	
	while (p && p != orig)
		p = p->next ;
		
	if (! p)
		err_sys("Cf_insert_arc :: Couldn't find origin node with label %d !", orig->label) ;
	
	arc->next = p->f_arc;
	p->f_arc = arc;
/*	a = p->f_arc ;

	//head insert

	p->f_arc = (CF_PTR_arc)malloc(sizeof(CF_arc)) ;
	if (!guard)
		p->f_arc->guard = NULL ;
	else
	{
		p->f_arc->guard = (char *) malloc(sizeof(char) * (strlen(guard)+1)) ;
		strcpy(p->f_arc->guard, guard) ;
	}
	if (!action)
		p->f_arc->action = NULL ;
	else
	{
		p->f_arc->action = (char *) malloc(sizeof(char) * (strlen(action)+1)) ;
		strcpy(p->f_arc->action, action) ;
	}
	if (!message)
		p->f_arc->msg = NULL ;
	else
	{
		p->f_arc->msg = (char *) malloc(sizeof(char) * (strlen(message)+1)) ;
		strcpy(p->f_arc->msg, message) ;
	}

	
	p->f_arc->extr = extr ;
	p->f_arc->next = a ;	
*/
}

CF_PTR_node CF_get_succ(CF_PTR_node node, int i)
{
	CF_PTR_arc a = CF_get_succ_arc(node, i) ;
	return a?a->extr:NULL ;
}

CF_PTR_arc CF_get_succ_arc(CF_PTR_node node, int i)
{
	CF_PTR_arc a = node->f_arc ;
	int c = 1 ;
	
	while (a && c < i)
	{
		a = a->next ;
		++c ;
	}
	
	//here, a migth be equal to null, which means there's no ith successor.
	return a ;
}


void CF_get_label(CF_PTR_arc arc, char ** guard, char ** action, char ** msg)
{
	*guard = arc->guard ;
	*action = arc->action ;
	*msg = arc->msg ;
}

/*creates an arc, given its extremity, its guard, its action and its message*/
CF_PTR_arc CF_create_arc(CF_PTR_node extr, char * guard, char * action, char * message)
{
	CF_PTR_arc arc = (CF_PTR_arc)malloc(sizeof(CF_arc));
	if (guard == NULL) {
		arc->guard = NULL;
	}
	else {
		arc->guard = (char *)malloc(sizeof(char) * (strlen(guard)+1));
		strcpy(arc->guard,guard);
	}
	if (action == NULL) {
		arc->action = NULL;
	}
	else {
		arc->action = (char *)malloc(sizeof(char) * (strlen(action)+1));
		strcpy(arc->action,action);
	}
	if (message == NULL) {
		arc->msg = NULL;
	}
	else {
		arc->msg = (char *)malloc(sizeof(char) * (strlen(message)+1));
		strcpy(arc->msg,message);
	}
	arc->info = NULL;
	arc->next = NULL;
	arc->extr = extr;
	arc->mark = 0 ;
	return arc;	
}

void CF_set_label_arc(CF_PTR_arc arc, char * guard, char * action, char * msg)
{
        if (!guard)
                arc->guard = NULL ;
        else
        {
                arc->guard = (char *) malloc(sizeof(char) * (strlen(guard)+1)) ;
                strcpy(arc->guard, guard) ;
        }
	if (!action)
                arc->action = NULL ;
        else
        {
                arc->action = (char *) malloc(sizeof(char) * (strlen(action)+1)) ;
                strcpy(arc->action, action) ;
        }
        if (!msg)
                arc->msg = NULL ;
        else
        {
                arc->msg = (char *) malloc(sizeof(char) * (strlen(msg)+1)) ;
                strcpy(arc->msg, msg) ;
        }
						
}

void CF_set_extr(CF_PTR_arc arc, CF_PTR_node extr)
{
	arc->extr = extr;
}

/*Adds the given arc to the list of outgoing arcs of the given node*/
void CF_set_arc(CF_PTR_node node, CF_PTR_arc arc)
{
	arc->next = node->f_arc;
	node->f_arc = arc;
}

CF_PTR_node CF_get_extr(CF_PTR_arc arc)
{
	return arc->extr ;
}

void CF_set_info_arc(CF_PTR_arc arc, void * info){
	arc->info = info;
}


void CF_dispose(CF_PTR_graph g)
{
	// free the graph
}

/*Merge nodes d and s. The resulting node will have all the successors of d and s and all the predecessors of d*/
void CF_merge_nodes(CF_PTR_node d, CF_PTR_node s)
{
	// We simply have to add all the outgoing arcs of s to
	// the list of arcs of d
	
	CF_PTR_arc a = d->f_arc ;
	if (!a) // no outgoing arc, the head of d's list becomes the head of s's lists
	{
		d->f_arc = s->f_arc ;
	}
	else // otherwise, we add d's list of arcs oat the end of s's
	{
		while (a->next)
			a = a->next ;
		// a is now the last arc of the list
		a->next = s->f_arc ;
	}
}

void CF_dump_graph(CF_PTR_graph g)
{
	char * buf = (char *)malloc(5000*sizeof(char)) ;
	FILE * out ;
	
	CF_dump_node(g->first, buf, 1) ;
	CF_dump_node(g->first, buf, 2) ;

	out = fopen("out.dot", "a") ;
	fprintf(out,"digraph controlflow {size=\"8,11\"\n%s\n}\n\n",buf) ;
	//printf("digraph controlflow {\n%s\n}\n\n",buf) ;
	fclose(out) ;
}

void CF_dump_node(CF_PTR_node n, char * buf, int pass)
{
	CF_PTR_arc a ;
	char * label =(char *)malloc(500*sizeof(char)) ;
	
	if (n && n->aux != pass)
	{
		//Let's add the node to the file
		if (pass == 1)
		{
			sprintf(label, "node%x [label=%s %s] ;\n", (int)n, (n->label==-1?"bidon":_label_table->table[n->label].name),
			(n->mark!=0 ? ", shape=diamond" : "")) ;
			strcat(buf, label) ;
		}
		
		//We make a recursive call on every successor
		++n->aux  ;
		a = n->f_arc ;
		while (a)
		{
			// let's add the arc	
			if (pass == 2 && a->mark == 0)
			{
				sprintf(label, "node%x -> node%x [label=\"%s, %s : %s\"] ;\n", (int)n, (int)a->extr, a->guard, a->action, a->msg) ;
				strcat(buf, label) ;	
				a->mark = 1 ;
			}
			CF_dump_node(a->extr,buf, pass) ;
			a = a->next ;
		}
	}

}
