/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
   
/***********************************************

Control graph data structure

************************************************/
#ifndef _controlflow_h
#define _controlflow_h

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "tbsymbol.h"
#include "error.h"
#include "stringmanip.h"

// These are the values of the mark field of CF_arc:
// _RDV_NODE => node whos outgoing arc is a rendez-vous with parameters
// _ACCEPT_NODE => same with accept
// _ASSIGN_NODE => the outgoing arc is an assignation
#define _NORMAL_NODE 0
#define _RDV_NODE 1 
#define _ASSIGN_NODE 2
#define _ACCEPT_NODE 3
#define _IF_NODE 4
#define _WHILE_NODE 5
#define _CHOICE_NODE 6

extern  T_tbsymbol * _label_table ;

struct s_CF_node ;

//One arc of the graph

typedef struct s_CF_arc {
	char * guard ;
	char * action ;
	char * msg ;
	struct s_CF_arc * next; // next arc in the list of arcs
	struct s_CF_node * extr ; // extremity of the arc
	int mark ;
	void * info;
} CF_arc ;

typedef CF_arc * CF_PTR_arc  ;

// One node of the graph
typedef struct s_CF_node {
	int label ;
	CF_PTR_arc f_arc ;  // first outgoing arc 
	void * info ;
	struct s_CF_node * next ;
	int mark ; // helps to find the "special nodes" 
	int aux ;
} CF_node ;

typedef CF_node * CF_PTR_node ;

// The control flow graph
typedef struct {
	CF_PTR_node head ; // the head of the list
	CF_PTR_node first ; // the first node in the execution flow
} CF_graph ;


typedef CF_graph * CF_PTR_graph ;

void CF_create_graph (CF_PTR_graph * graph) ;

void CF_insert_node(CF_PTR_graph graph, CF_PTR_node node) ;
CF_PTR_node CF_create_node(int label, void * info, int mark) ;
CF_PTR_node CF_get_node(CF_PTR_graph graph, int label) ;
void CF_set_info(CF_PTR_node node, void * info, int mark) ;
void CF_set_label_node(CF_PTR_node node, int label) ;
CF_PTR_node CF_get_succ_node(CF_PTR_node node, int i) ;
CF_PTR_arc CF_get_succ_arc(CF_PTR_node node, int i) ;
void * CF_get_info(CF_PTR_node node) ;


CF_PTR_arc CF_create_arc(CF_PTR_node extr, char * guard, char * action, char * message);

void CF_get_label(CF_PTR_arc arc, char ** guard, char ** action, char ** msg) ;
void CF_insert_arc(CF_PTR_graph graph, CF_PTR_node orig, CF_PTR_arc arc) ;
void CF_set_label_arc(CF_PTR_arc arc, char * guard, char * action, char * msg);
void CF_set_extr(CF_PTR_arc arc, CF_PTR_node extr);
void CF_set_arc(CF_PTR_node node, CF_PTR_arc arc) ;
CF_PTR_node CF_get_extr(CF_PTR_arc) ;
void CF_set_info_arc(CF_PTR_arc arc, void * info);

void CF_dump_node(CF_PTR_node n, char * buf, int pass) ;
void CF_dump_graph(CF_PTR_graph g) ;
void CF_merge_nodes(CF_PTR_node d, CF_PTR_node s) ;
#endif
