/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
   


#include "CBP.h"
#include "globalmachine.h"
#include "localvaluation.h"

int fakeNodeNum=0 ; /* Use by LM_get_state_name to give a name to the fake nodes */


/* This function simply creates a local machine with n local state */
PTR_local_machine LM_create_machine (int n)
{
	PTR_local_machine l = (PTR_local_machine)malloc(sizeof(local_machine)) ;
	l->list = NULL ;
	l->nLocVar = n ;
	l->init_state = NULL ;
	return l ;
}

/* This function creates a local state corresponding to node n in
   the CFG, and valuation v of the local variables (there are
   exactly nVar local variables)*/
LM_PTR_state LM_create_state (CF_PTR_node n, unsigned * v, int nVar) 
{
	int i ;
	LM_PTR_state s = (LM_PTR_state)malloc(sizeof(LM_state)) ;
	s->cfNode = n ;
	s->val = (unsigned int *)malloc(nVar*sizeof(unsigned int)) ;
	for (i=0; i<nVar; ++i)
		s->val[i] = v[i] ;
	
	s->next = NULL ;
	s->firstTrans = NULL ;
	return s ;
}

/* This function adds the already constructed local state s to the
   list of local states of local machine l.
   Warning ! The state is assumed NOT to exist in the list !*/
void LM_add_state (PTR_local_machine l, LM_PTR_state s) 
{
	LM_PTR_state_list p = l->list ;
	LM_PTR_state_list q = NULL ;
	LM_PTR_state_list r ;

	if (p == NULL) /* This is the first state to be inserted */
	{
		l->list = (LM_PTR_state_list)malloc(sizeof(LM_state_list)) ;
		l->list->next = NULL ;
		l->list->cfNode = s->cfNode;
		l->list->first = s ;
		s->next = NULL;
	}
	else
	{
		while ((p != NULL) && (s->cfNode > p->cfNode))
		{			
			q = p ;
			p = p->next ;
		}

		if (q == NULL) /* No state have been inserted for this CFG node yet -> head insert */
		{
			r = (LM_PTR_state_list)malloc(sizeof(LM_state_list)) ;
			r->next = p ;
			l->list = r ;
			r->cfNode = s->cfNode ;
			r->first = s ;
			s->next = NULL;
		}
		else if (p == NULL) {
                        /* thus, we have to insert a new list node between p and q */
                        r = (LM_PTR_state_list)malloc(sizeof(LM_state_list)) ;
                        r->next = NULL ;
                        q->next = r ;
                        r->first = s ;
                        s->next = NULL;
                        r->cfNode = s->cfNode ;

		}
		else if (s->cfNode == p->cfNode) /* Some valuations already exist for this CFG node */
		{
			/* We just insert s in the head of p->first */
			s->next = p->first ;
			p->first = s ;
		}
		else /* s->cfNode < p->cfNode */
		{
			/* thus, we have to insert a new list node between p and q */
			r = (LM_PTR_state_list)malloc(sizeof(LM_state_list)) ;
			r->next = p ;
			q->next = r ;
			r->first = s ;
			s->next = NULL;
			r->cfNode = s->cfNode ;
		}
	}
}

/* This function looks for the state corresponding to CFG node n and valuation v
   in the list of the local machine l. It returns a pointer to this state. */
LM_PTR_state LM_get_state (PTR_local_machine l, CF_PTR_node n, unsigned * v) 
{
	LM_PTR_state_list p = l->list ;
	LM_PTR_state q ;
	int i, nVar = l->nLocVar ;
	
	while (p && p->cfNode != n)
	{
		p = p->next ;
	}	
	
	if (p == NULL)
		return NULL ;
	else
	{
		/* We have found the sub-list corresponding to n. Now let's look at the valuation of the variables */
		q = p->first ;
		while (q)
		{
			for (i = 0; i < nVar && q->val[i] == v[i]; ++i)
				;
			
			if (i == nVar) /* It's the right valuation */
				break ;
				
			q = q->next ;
		}
	}
	
	return q ; /* might be null if the valuation hasn't been entered yet */
}


/* This function creates and adds a transition going from local sate o to local state e.
   The guard, action and label are as specified.*/
LM_PTR_trans LM_add_trans (LM_PTR_state o, LM_PTR_state e, char * action, char * guard, char * label) 
{
	LM_PTR_trans t = (LM_PTR_trans)malloc(sizeof(LM_trans)) ;
	
	/* Let's set the fields of this transition */

	/*precieuse : I modify to make the code uniform according to the rest (a copy of the strings are kept and not the strings) */
	t->extr = e ;

	if (!guard)
                t->guard = NULL ;
        else
        {
                t->guard = (char *) malloc(sizeof(char) * (strlen(guard)+1)) ;
                strcpy(t->guard, guard) ;
        }

//	t->guard = guard ;

	if (!action)
                t->action = NULL ;
        else
        {
                t->action = (char *) malloc(sizeof(char) * (strlen(action)+1)) ;
                strcpy(t->action, action) ;
        }

//	t->action = action ;

	if (!label)
                t->label = NULL ;
        else
        {
                t->label = (char *) malloc(sizeof(char) * (strlen(label)+1)) ;
                strcpy(t->label, label) ;
        }

//	t->label = label ;
	
	/* Now we head-insert it in the list of outgoing arcs of o */
	
	t->next = o->firstTrans ;
	o->firstTrans = t ;
	
	return t ;

}


PTR_global_machine GM_generate(void)
{
	int i, j, k ;
	int mainfound = 0 ;
	T_tbsymbol_entry e ;
	ThreadInfo * ti ;
	LM_PTR_state p ;
	LM_PTR_state first_state ;
	LM_PTR_state initial_state;
	char * buf = (char*)malloc(256*sizeof(char)) ;
	unsigned int * val ;
	
	PTR_global_machine G = (PTR_global_machine)malloc(sizeof(global_machine)) ;
	
	PTR_local_machine  L ; 
	
	LM_PTR_set_of_states WL1 = NULL, WL2 = NULL ;

	buf[0] = 0 ;
#ifdef __DEBUG

	for (i=0; i<_thread_table->nbr_entries; ++i)
	{
		e=_thread_table->table[i] ;
		ti = (ThreadInfo *)e.info ;
		if (!ti)
		{
			err_sys("Thread type %s has been declared but I couldn't find its body ! Giving up !", e.name) ;
		}
		CF_dump_graph(ti->cfg) ;
	}
#endif

	/* Initialisation of the GM */
	G->first = NULL ;
	G->nGlobVar = _glob_var_table->nbr_entries ;
	
	for (i=0; i<_thread_table->nbr_entries; ++i)
	{
		e=_thread_table->table[i] ;
		ti = (ThreadInfo *)e.info ;
		if (!ti)
		{
			err_sys("Thread type %s has been declared but I couldn't find its body ! Giving up !", e.name) ;
		}
	
		printf ("\tGM_generate : Handling local machine for thread %s !\n", e.name) ;
		if (strcmp(e.name, "main") == 0) 
		{
			/* This is the main thread, which means we have no start for it. Therefore, we must
			   abitrarily fix its initial local values. We choose to put them to true. */
			   
			mainfound = 1 ;
			
			/* Let's create a local machine */
			if (ti->vartable == NULL)
				L = LM_create_machine(0) ;
			else
				L = LM_create_machine(ti->vartable->nbr_entries) ;
			
			/* The first state corresponds to the input node of the cfg with valuation T...T */
			val = (int*)malloc(L->nLocVar*sizeof(int)) ;
			for( k=0; k<L->nLocVar; ++k)
				val[k] = 1 ;
				
			first_state = LM_create_state(ti->cfg->first, val, L->nLocVar) ;	
			LM_add_state(L, first_state) ;
			
			/* Let's record it as the initial state */ 
			
			L->init_state = first_state ;
			
			/* It's gonna be our first working frontier... */
			
			LM_insert_state_into_set(&WL1, first_state);
		}
		else /* In the case it is a normal thread */
		{
			
			if ((ti->vartable != NULL && ti->initStates->nb_variable != ti->vartable->nbr_entries) || (ti->vartable == NULL
					&& ti->initStates->nb_variable != 0))
				err_sys("Error: the starts for thread %s have %d variables, but %d variables have been declared for it !\n",
				e.name,  ti->initStates->nb_variable, ti->vartable->nbr_entries) ;
			
			L = LM_create_machine (ti->initStates->nb_variable) ;

			/* We have to create a first state to allow us to synchronize with the starts */
		
			val = (int*)malloc(L->nLocVar*sizeof(int)) ;
			for( k=0; k<L->nLocVar; ++k)
				val[k] = 0 ;
		
			first_state = LM_create_state(NULL, val, L->nLocVar) ;
			LM_add_state(L, first_state) ;
	
			/* Let's have a look at all the intial states computed thanks to precieuse's wonderful functions*/
			p = ti->initStates->first_state ;

			printf("\tGM_generate : Generating initial states...") ;
			
			if (L->nLocVar == 0) /* If this thread has no local variable, we have to create a state
						corresponding to the first node of the cfg, but w/o valuation ! */
			{
				/* We can reuse val for the valuation: it's empty anyway ! */
				initial_state = LM_create_state(ti->cfg->first, val, L->nLocVar) ; 
				
				LM_add_state (L, initial_state) ;
				
				/* This sole state will be our first working frontier */
				LM_insert_state_into_set(&WL1, initial_state) ;
				
				/* We finish by adding a transition from first_state to initial_state
				   in order to let the LM synchronize with the start */
				sprintf(buf, "_start_%s?", e.name) ;
				
				LM_add_trans (first_state, initial_state, "", "", buf) ;
			}
			else
			{
				while (p) /* If L->nLocVar == 0, we are sure that p ==  NULL */
				{		
					/* First, we construct the state to add*/
					initial_state = LM_create_state(ti->cfg->first, p->val, L->nLocVar);
					LM_add_state (L, initial_state) ;
			
					/* We put it in the first work-list */
					LM_insert_state_into_set(&WL1, initial_state);
				
					/* And we add a transition from first_state to intial_state that makes an accept 
					   with = label _state_'thread name''valuation of the local variable in initial_state'? */
				
					sprintf(buf, "_start_%s", e.name) ;
				
					for (k=0; k<L->nLocVar; ++k)
					{
						if (p->val[k] == 1)
							strcat(buf, "T") ;
						else
							strcat(buf, "F") ;
					}	
					strcat(buf, "?") ;
				
					LM_add_trans (first_state, initial_state, "", "", buf) ;
			
					p = p->next;
				}
			}
			
			printf(" done !\n\n") ;
		}
		
		
		
		/* Now, we can compute the other states (fix-point) */
			
		printf ("\tGM_generate : Starting fix-point computation...\n") ;
		j =  0 ;
		printf("\t\titeration   %d", j) ;
		/* Now, the fix point */
		while (WL1 != NULL)
		{
			++ j ;
			printf("\b") ;
			if (j > 10)
				printf("\b") ;
			if (j > 100)
				printf("\b") ;
			printf("%d", j) ;
			WL2 = LM_post(WL1, L) ;
			/* Free WL1 */
			WL1 = WL2 ;
		}
		printf(" done !\n") ;
		
		/* The LM is constructed */
		ti->locMach = L ;
		L->name = (char *)malloc((strlen(e.name)+1)*sizeof(char)) ;
		strcpy(L->name, e.name) ;
		
		/* Let's put it in the list of the GM */
		L->next = G->first ;
		G->first = L ;
		
	}
	
	if (mainfound == 0)
	{
		printf("\tGM_generate: Warning ! I was expecting a 'main' thread and I haven't found it !! Going on anyway...\n") ;
	}
	
	return G ;
}

/* This function  writes to f the decription of the local machine L in the appropriate format. name is the name of the local machine. 
*/
void LM_write_to_file(FILE * f, PTR_local_machine L)
{
	LM_PTR_state_list p ;
	LM_PTR_state q ;
	LM_PTR_trans t ;
#ifdef __DOT
	int i ;
	FILE * dot ;
	char * buf = (char*)malloc(100*sizeof(char)) ;
	char * temp ;
#endif
	int n ;
	int numStates = 0, numTrans = 0 ;
	char * sName;
	char * dName ;
	char * transBuf = (char *)malloc(10000*sizeof(char)) ; /* To store the transitions */
	
	transBuf[0] = 0 ;
	fprintf(f, "Local Machine %s ;\n", L->name) ;
	fprintf(f, "begin\n\t States\n\t\t" );
	
	/* We must give the list of all the states. Let's look for them in the list L->list */
	p = L->list ; /* The head of the high-level list */
	n = 0 ; /* Do we have to put a comma ? */
#ifdef __DOT
	sprintf(buf, "%s.dot", L->name) ;
	dot = fopen(buf, "w+") ;
	fprintf(dot, "digraph local_machine_%s {size=\"8,12\"\n", L->name) ;
#endif
	
	while (p != NULL)
	{
	
		q = p->first ; /* The first state for a given cfg node */
		while (q != NULL)
		{
			/* What's its name ? */
			sName = LM_get_state_name(q, L) ;	
			++numStates ;
			
			if (n != 0) /* Not the first state to be written => we can put a ','*/
				fprintf(f, ", ") ;
			else
				n = 1 ;
			fprintf(f, "%s", sName) ;
			
#ifdef __DOT
			buf[0] = 0 ;	
			strcat(buf, "(") ;
			for (i = 0; i<L->nLocVar; ++i)
			{
				if (q->val[i] == 1)
					strcat(buf, "true, ") ;
				else
					strcat(buf, "false, ") ;
			}
			i = strlen(buf) ;
			buf[i-2]=')' ;
			buf[i-1] = 0 ;
			if (q->cfNode != NULL)
				temp = tbsymbol_get_num(_label_table, q->cfNode->label)->name ;
			else
				temp = "idle" ;
			fprintf(dot, "node%x [label=\"%s\\n %s\"] ;\n", (int)q, temp, buf) ;
#endif			
			
			/* Now let's have a look a this state's transitions.... */
			
			t = q->firstTrans ;
			while (t != NULL)
			{
				strcat(transBuf, "\t\t") ;
				strcat(transBuf, sName) ;
				strcat(transBuf, " - ") ;
				strcat(transBuf, t->label) ;
				strcat(transBuf, " -> ") ;
				
				/* We need the name of the destination state */
				
				dName = LM_get_state_name(t->extr, L) ;
				strcat(transBuf, dName) ;
				if ((t->guard[0] != 0) || (t->action[0] != 0)) /* Are these strings empty ? */
				{
					strcat(transBuf, " : ") ;
					if (t->guard[0] != 0)
					{
						strcat(transBuf, t->guard) ;
						if (t->action[0] != 0)
							strcat(transBuf, _DELIM_) ;
					}
					if (t->action[0] != 0)
						strcat(transBuf, t->action) ;
				}
				strcat(transBuf, " ;\n") ;
				
				
				free(dName) ;
				++numTrans ;
#ifdef __DOT
				if (strlen(t->action) ==0)
				{
					if (strlen(t->guard) == 0)
						fprintf(dot, "node%x -> node%x [label=\" %s \"] ;\n", (int)q, (int)t->extr, t->label) ;
					else
						fprintf(dot, "node%x -> node%x [label=\" %s : %s\"];\n", (int)q, (int)t->extr, t->label, t->guard) ;
				}
				else
				{
					if (strlen(t->guard) == 0)
						fprintf(dot, "node%x -> node%x [label=\" %s : true -> %s\"] ;\n", (int)q, (int)t->extr, t->label, t->action) ;
					else
						fprintf(dot, "node%x -> node%x [label=\" %s : %s -> %s\"] ;\n", (int)q, (int)t->extr, t->label, t->guard, t->action) ;
				}
#endif				
				t = t->next ;
			}
			
			
			q = q->next ;
			free(sName) ;
		}	
		
		p = p->next ;
	}
#ifdef __DOT
	fprintf(dot, "}\n\n") ;
	fclose(dot) ;
#endif
	

	fprintf(f, " ;\n\n\t Transitions\n") ;
	fprintf(f, "%s", transBuf) ;
	if (strcmp(L->name, "main") == 0)
		fprintf(f, "\n\t Initial State\n\t\t %s ;\n", LM_get_state_name(L->init_state, L)) ;
	else
		fprintf(f, "\n\t Initial State\n\t\t %s_init_ ;\n", L->name) ; /* By convention. See function LM_get_state_name() */
	fprintf(f, "end\n\n") ;
	
	printf("\tLocal machine %s has been generated with %d states and %d transitions.\n",L->name, numStates, numTrans) ;
}

/* This function returns a string that uniquely identifies the LM state s (which belongs to a local machine L) */
char * LM_get_state_name(LM_PTR_state s,  PTR_local_machine L)
{
	
	char * sName = (char *)malloc((100+L->nLocVar)*sizeof(char)) ;
	char * buf = (char*) malloc((L->nLocVar+1)*sizeof(char)) ;
	char * buf2 = (char*)malloc(15*sizeof(char)) ;
	int j ;
	T_PTR_tbsymbol_entry e ;
	buf[0] = 0 ;
	buf2[0] = 0 ;
	
	/* The name of the state will be: the concatenation of the valuation of the local variables, 
	the LM name and the cfg's label. If the cfg node is a "fake" node (it has label = -1), we assume
	its label is _FAKE_LMname_i_, where i is a continuously incremented variable. If the cfg node is NULL,
	then it is the very first state of the local machine, and we give it a special name _init_ */
	
	/* Let's first evaluate the local variables */
	
	for (j=0; j<L->nLocVar; ++j)
	{
		if (s->val[j] == 0) 
			strcat(buf, "F") ;
		else
			strcat(buf, "T") ;
	}
	
	/* Now we can write the name of the state */
	if(s->cfNode == NULL)
	{
		sprintf(sName, "%s_init_", L->name) ;
	}
	else if (s->cfNode->label != -1)
	{
		e = tbsymbol_get_num(_label_table, s->cfNode->label);
		if (L->nLocVar != 0)
			sprintf(sName, "%s@%s_%s", buf,L->name, e->name) ;
		else
			sprintf(sName, "%s_%s", L->name, e->name) ;
	}
	else
	{
		sprintf(buf2, "_FAKE_%d_", fakeNodeNum ) ;
		if (L->nLocVar != 0)
			sprintf(sName, "%s@%s_%s",buf,L->name, buf2) ;
		else
			sprintf(sName, "%s_%s", L->name, buf2) ;	
		++fakeNodeNum ;
		/* Let's record this name in the label symbol table */
		s->cfNode->label = tbsymbol_insert_num(_label_table, buf2 );
	}
	
	return sName ;
}

/* This function writes the whole GM 'G' to the file 'f'. The name of the GM is 'name' */
void GM_write_to_file(PTR_global_machine G, FILE * f, char * name)
{
	PTR_local_machine L ;
	PTR_msgList p ;
	int i ;
	
	fprintf(f, "Global Machine GM:\n") ;
	fprintf(f, "%% Generated from %s\n\n", name) ;
	fprintf(f, "begin\n\tBool\n\t\t") ;
	
	/* We find the global boolean variables in the _glob_var_table symbol table */
	if (_glob_var_table->nbr_entries != 0)
	{
		for (i = 0; i< _glob_var_table->nbr_entries - 1; ++i)
		{
			fprintf(f, "%s, ", _glob_var_table->table[i].name) ;
		}
			fprintf(f, "%s ", _glob_var_table->table[i].name) ;
	}
	
	if ( _lock_table->nbr_entries != 0)
		fprintf(f, ", ") ;
	
	/* We find the locks in the _lock_table symbol table */
	if (_lock_table->nbr_entries != 0)
	{
		for (i = 0; i< _lock_table->nbr_entries - 1; ++i)
		{
			fprintf(f, "%s, ", _lock_table->table[i].name) ;
		}
		fprintf(f, "%s ", _lock_table->table[i].name) ;
	}
	fprintf(f, " ;\n") ;
	
	
	fprintf(f, "\n\tActions\n\t\t") ;
	
	/*The actions can be found in the mList list */
	
	p = mList ;
	
	while (p)
	{
		fprintf(f, "%s, ", p->message) ;
		p = p->next ;
	}
	
	/* And we finish by all the actions used to simulate the keywords of the language */
	
	fprintf(f, "%s ;\n\n", _COMMON_ACTIONS_) ;

	/* Now let's generate all the local machines */
	
	L = G->first ;
	while (L)
	{
		LM_write_to_file(f, L) ;
		L = L->next ;
	}
	
	fprintf(f, "Initial Global State\n") ;
	
	L = G->first ;
	while (L)
	{
		fprintf(f, "\t#%s = any ; %%replace by a value if necessary !\n", L->name) ;
		L = L->next ;
	}
	
	fprintf(f, "\nUnsafe States\n") ;
	fprintf(f, "\t%% Put here your property (formula over the number of local machine in a given local state)\n\n") ;
	fprintf(f, "end\n\n") ;
	fprintf(f, "%% Generated by CBP2GM v %s \n\n", _version_) ;
}


PTR_LM_initial_states LM_get_local_states(LM_list_initial_states list, char *local_machine) {
	PTR_LM_initial_states sol;
	PTR_LM_initial_states ptr;
	
	ptr = (PTR_LM_initial_states)list.first_element;
	while (ptr != NULL) {
		if (strcmp(ptr->thread,local_machine) == 0) {
			sol = ptr;
			ptr = NULL;
		}
		else {
			ptr = ptr->next;
		}
			
	}
	return sol;
}

/*return 0 if the states are differents
  return 1 if the states are the same*/
int LM_same_local_state(LM_state state1, LM_state state2,int nb_variable) {
	int sol = 1;
	int i;
	
	if (state1.cfNode == state2.cfNode) {
		for (i=0;((i < nb_variable) && sol == 1);i++) {
			if (state1.val[i] != state2.val[i])
				sol = 0;
		}
	}
	else
		sol = 0;
	return sol;
}

/*return 0 if it is not in the list
  return 1 if it is in the list*/
int LM_contain_state(LM_PTR_state state, PTR_LM_initial_states list) {
	LM_PTR_state ptr = list->first_state;
	int sol = 0;
	while (ptr != NULL) {
		if (LM_same_local_state(*ptr,*state,list->nb_variable) == 1) {
			sol = 1;
			ptr = NULL;
		}
		else
			ptr = ptr->next;
	}
	return sol;
}


PTR_LM_initial_states LM_create_initial_states(char * name) {
	PTR_LM_initial_states sol = (PTR_LM_initial_states)malloc(sizeof(LM_initial_states));
	sol->thread =(char *)malloc(sizeof(char) * strlen(name));
	strcpy(sol->thread,name);
	sol->first_state = NULL;
	sol->next = NULL;
	return sol;
}

void LM_insert_local_state(LM_PTR_state state, PTR_LM_initial_states list) {
	state->next = list->first_state;
	list->first_state = state;
}

void LM_insert_list_local_states(PTR_LM_list_initial_states list, char *name_thread) {
	PTR_LM_initial_states new_list = LM_create_initial_states(name_thread);
	new_list->next = list->first_element;
	list->first_element = new_list;
}

void create_list_local_states(PTR_LM_list_initial_states * list) {
	*list =(PTR_LM_list_initial_states) malloc (sizeof(LM_list_initial_states));
	(*list)->first_element = NULL;
}

//computes the successor operation for a normal local state and a control flow arc
//the function returns the successor state only if it is a new local state
LM_PTR_state LM_post_normal_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l) {
	LM_PTR_state successor_state;
	LM_PTR_trans trans;

	successor_state = LM_get_state (l,control_flow_arc.extr,state->val);
	if (successor_state == NULL) {
		successor_state = LM_create_state(control_flow_arc.extr,state->val,l->nLocVar);
		LM_add_state (l, successor_state);
		trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
	}
	else {
		trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
		successor_state = NULL;
	}
	return successor_state;
}

//computes all the successors of a normal node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_normal_node(LM_PTR_state state, PTR_local_machine l) {
	CF_PTR_arc control_flow_arc;
	LM_set_of_states * list_of_states = NULL;
	LM_PTR_state successor_state;
	
	control_flow_arc = state->cfNode->f_arc;
	while (control_flow_arc != NULL) {
		successor_state = LM_post_normal_node_arc(state,*control_flow_arc,l);
		if (successor_state != NULL)
			LM_insert_state_into_set(&list_of_states,successor_state); /* to check */
		control_flow_arc = control_flow_arc->next;
	}
	return list_of_states;
}


//inserts a state into a set of states
void LM_insert_state_into_set(LM_set_of_states **set,LM_PTR_state state) {
	LM_set_of_states * new_element = (LM_set_of_states *) malloc (sizeof(LM_set_of_states));
	new_element->state = state;
	new_element->next = *set;
	*set = new_element;
}

//merges two set of states
void LM_merge_set_of_states(LM_set_of_states *set1, LM_set_of_states *set2) {
	LM_set_of_states *ptr;

	ptr = set1;
	while (ptr->next != NULL) 
		ptr = ptr->next;
	ptr->next = set2;
}

//computes the successor operation for a if local state and a control flow arc
//the function returns the successor state only if it is a new local state
LM_PTR_state LM_post_if_node_arc(LM_state * state, CF_arc control_flow_arc, PTR_local_machine l) {
	        LM_PTR_state successor_state;
	        LM_PTR_trans trans;
		int *test;

		if (control_flow_arc.info == NULL) {/* No condition on the arc */
			successor_state = LM_post_normal_node_arc(state,control_flow_arc,l);
		}
		else {	/* there is a condition, we have to evaluate it */
				
			test = transform_expression((ExpressionRet *)control_flow_arc.info, l->nLocVar);		
			if (test_value(state->val,test,l->nLocVar) == 1) { /* If the local values in this state correspond to the
			valuation requested by the if */
				successor_state = LM_get_state (l,control_flow_arc.extr,state->val);
		        	if (successor_state == NULL) {/* If this state doesn't exist in the LM yet, we have to create it */
	        	        	successor_state = LM_create_state(control_flow_arc.extr,state->val,l->nLocVar);
		                	LM_add_state (l, successor_state);
			                trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
			        }
	        		else {/* Otherwise, we simply create the transition */
	                		trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
			                successor_state = NULL;
			        }
			}
			else 
				successor_state = NULL;
			free(test);
		}
	        return successor_state;
}

//computes all the successors of a if node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_if_node(LM_state * state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL;
        LM_PTR_state successor_state;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) {
                successor_state = LM_post_if_node_arc(state,*control_flow_arc,l);
                if (successor_state != NULL)
                        LM_insert_state_into_set(&list_of_states,successor_state); /* to check */
	        control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;
}

//computes the successor operation for a while local state and a control flow arc
//the function returns the successor state only if it is a new local state
LM_PTR_state LM_post_while_node_arc(LM_state * state, CF_arc control_flow_arc, PTR_local_machine l) {
	LM_PTR_state successor_state;
        LM_PTR_trans trans;
        int *test = transform_expression((ExpressionRet *)state->cfNode->info, l->nLocVar);

        if (((test_value(state->val,test,l->nLocVar) == 1) && (strcmp(control_flow_arc.msg,"while") == 0))
		|| ((test_value(state->val,test,l->nLocVar) == 0) && (strcmp(control_flow_arc.msg,"endwhile") == 0))){
		successor_state = LM_get_state (l,control_flow_arc.extr,state->val);
                if (successor_state == NULL) {
			successor_state = LM_create_state(control_flow_arc.extr,state->val,l->nLocVar);
		        LM_add_state (l, successor_state);
                        trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
                }
                else {
                        trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
                        successor_state = NULL;
                }
        }
        else 
		successor_state = NULL;
        free(test);
        return successor_state;
}

//computes all the successors of a if node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_while_node(LM_state * state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL; 
        LM_PTR_state successor_state;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) { 
                successor_state = LM_post_while_node_arc(state,*control_flow_arc,l);
                if (successor_state != NULL)
                        LM_insert_state_into_set(&list_of_states,successor_state); /* to check */
                control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;
}

/* This function is called each time we encounter a rendez-vous node in the cfg. When doing a RDV,
   we expect the corresponding accept to exist. For simplicity reasons, it will synchronize on
   all the possible values for the local variables. Therefore, we can already safely generate all the messages
   of the form msg'all possible valuation of the loc. var.', and add them to the list mList*/

void LM_generate_rdv_msg(char * msg, int n)
{
	char * buf = (char*)malloc(sizeof(char)*(n+strlen(msg)+1)) ;
	char * val = (char*)malloc(sizeof(char)*(n+1)) ;
	int i, j = n-1 ;
	for (i=0; i<n; ++i)
		val[i]= 'F' ;
	val[n] = 0 ;
	
	while (1)
	{
		/* Let's insert the message in the list */
		strcpy(buf, msg) ;
		strcat(buf, val) ;
		msgList_insert(buf, &mList) ;
		
		/*printf("LM_generate_rdv_msg: adding %s to the table\n", buf) ;*/
		
		/* Now we generate another one */
		while (val[j] == 'T' && j >= 0)
		{
			val[j] = 'F' ;
			--j ;
		}
		if (j < 0) /* Finished ! */
			break ;
		val[j] = 'T' ;
		j = n-1 ;
	}
}

//computes the successor operation for a rendez-vous local state and a control flow arc
//the function returns the successor state only if it is a new local state
LM_PTR_state LM_post_rdv_node_arc(LM_state * state, CF_arc control_flow_arc, PTR_local_machine l) {
        LM_PTR_state successor_state;
        LM_PTR_trans trans;
	ExpListRet *ptr = (ExpListRet *) state->cfNode->info;
	int val;
	char *msg = (char *) malloc (sizeof(char) * (strlen(control_flow_arc.msg) + l->nLocVar + 2));

	//construction of the synchronization message
	strcpy(msg, control_flow_arc.msg) ;
	
	LM_generate_rdv_msg(msg, l->nLocVar) ;
	
	while (ptr != NULL) {
		if (ptr->exp->type == _LOCVAR)
		{
			val = state->val[ptr->exp->var];
			if (ptr->exp->not == 1) {
				if (val == _TRUE_VALUE)
					val = _FALSE_VALUE;
				else
					val = _TRUE_VALUE;
			}
		}
		else if(ptr->exp->type == _LITTRUE)
		{
			val = _TRUE_VALUE ;
		}
		else if(ptr->exp->type == _LITFALSE)
		{
			val = _FALSE_VALUE ;
		}
			
		if (val == _TRUE_VALUE)
			strcat(msg,"T") ;
		else
			strcat(msg,"F") ;
		ptr = ptr->next ;
	}
	
	strcat(msg, "!") ;
	
        successor_state = LM_get_state (l,control_flow_arc.extr,state->val);
        if (successor_state == NULL) {
        	successor_state = LM_create_state(control_flow_arc.extr,state->val,l->nLocVar);
                LM_add_state (l, successor_state);
                trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, msg);
        }
        else {
                trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, msg);
                successor_state = NULL;
        }
        return successor_state;
}

//computes all the successors of a rendez-vous node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_rdv_node(LM_state * state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL; 
        LM_PTR_state successor_state;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) {
                successor_state = LM_post_rdv_node_arc(state,*control_flow_arc,l);
                if (successor_state != NULL)
                        LM_insert_state_into_set(&list_of_states,successor_state); /* to check */
                control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;
}


/*This function removes the assignation from local variables to global variables from list, and returns a list containing these
assignations 
Side effect: the orderings of the lists are reversed*/
AssignRet *list_assign_local_to_global(AssignRet **list) {
	AssignRet *sol=NULL;
	AssignRet *rest=NULL;
	AssignRet *ptr = *list;
	AssignRet *nextptr;

	while(ptr != NULL) {
		nextptr = ptr->next;
		if ((ptr->varType == _GLOBVAR) && (ptr->exp->type == _LOCVAR)) {
			ptr->next = sol;
			sol = ptr;
		}
		else {
			ptr->next = rest;
			rest = ptr;
		}
		ptr = nextptr;
	}
	(*list) = rest;
	return sol;
}

/*This function removes the assignation from local variables to local variables from list, and returns a list containing these
assignations 
Side effect: the orderings of the lists are reversed*/
AssignRet *list_assign_local_to_local(AssignRet **list) {
	AssignRet *sol=NULL;
	AssignRet *rest=NULL;
	AssignRet *ptr=*list;
	AssignRet *nextptr;

	while(ptr != NULL) {
		nextptr = ptr->next;
		if ((ptr->varType == _LOCVAR) && (ptr->exp->type == _LOCVAR)) {
			ptr->next = sol;
			sol = ptr;
		}
		else {
			ptr->next = rest;
			rest = ptr;
		}
		ptr = nextptr;
	}
	(*list) = rest;
	return sol;
}

//return the assignation from literals to local variables
AssignRet *list_assign_literal_to_local(AssignRet **list) {
	AssignRet *sol=NULL;
	AssignRet *rest=NULL;
	AssignRet *ptr = *list;
	AssignRet *nextptr;

	while(ptr != NULL) {
		nextptr = ptr->next;
		if ((ptr->varType == _LOCVAR) && ((ptr->exp->type == _LITTRUE) || (ptr->exp->type == _LITFALSE))) {
                        ptr->next = sol;
			sol = ptr;
		}
		else {
			ptr->next = rest;
			rest = ptr;
		}
		ptr = nextptr;
	}
	(*list) = rest;
	return sol;
}
		
LM_PTR_state construct_succ_assign_arc(CF_arc arc, int* newvaluation, LM_state *state,PTR_local_machine l,
		LM_list_of_variable * listvariables,AssignRet * globtoloc,AssignRet * gv, char * _action){
	LM_PTR_trans trans;
	LM_PTR_state successor_state;
        LM_list_of_variable * ptr;
        AssignRet * ptr_assign;
        char *guard = (char *)malloc(sizeof(char) * 500);
        char *action = (char *)malloc(sizeof(char) * 500);
        char *varname;
	int val;

	//construction of the guard
        strcpy(guard, arc.guard);
        ptr = listvariables;
        while (ptr != NULL) {
        	varname = tbsymbol_get_num(_glob_var_table, ptr->varId)->name;
                if (ptr->value == _TRUE_VALUE) {
			add_to_guard(guard, varname, 0) ;
			/*
	                strcat(guard,varname);
                        strcat(guard, " ");*/
                }
                else {
			add_to_guard(guard, varname, 1) ;
			/*
                        strcat(guard,"not(");
                        strcat(guard,varname);
                        strcat(guard,") ");*/
                }
		ptr = ptr->next;
	}

	//construction of the action
	strcpy(action,_action);
	//for assignation from global to global
	ptr_assign = gv;
        while ( ptr_assign != NULL) {
        	varname = tbsymbol_get_num(_glob_var_table, ptr_assign->varId)->name;
                val = value_of_var(listvariables, ptr_assign->exp->var);
                if (ptr_assign->exp->not == 1) {
	                if (val == _TRUE_VALUE)
	                        val  = _FALSE_VALUE;
                        else
                                val = _TRUE_VALUE;
                }

                if (val == _TRUE_VALUE) {
			add_to_action(action, varname, _STRUE_) ;
			/*
			strcat(action,varname);
                        strcat(action,":= true ");*/
                }
                else {
			add_to_action(action, varname, _SFALSE_) ;
			/*
                        strcat(action,varname);
                        strcat(action,":= false ");*/
                }
		ptr_assign = ptr_assign->next;
	}

	//for assignation from global to local
	ptr_assign = globtoloc;
        while (ptr_assign != NULL) {
		val =value_of_var(listvariables, ptr_assign->exp->var);
	       	if (ptr_assign->exp->not == 1) {
			if (val == _TRUE_VALUE)
	        		val  = _FALSE_VALUE;
	               	else
        	               	val = _TRUE_VALUE;
	       	}
	       	newvaluation[ptr_assign->varId] = val;
		ptr_assign = ptr_assign->next;
	}

	//construction of the successor
	successor_state = LM_get_state(l,arc.extr,newvaluation);
	if (successor_state == NULL) {
                successor_state = LM_create_state(arc.extr,newvaluation,l->nLocVar);
                LM_add_state (l, successor_state);
                trans = LM_add_trans (state, successor_state,action,guard, arc.msg);
        }
        else {
                trans = LM_add_trans (state, successor_state,action,guard, arc.msg);
                successor_state = NULL;
        }
	return successor_state;			
}

/* no documentation available yet for this function... */
LM_PTR_set_of_states assign_succ(CF_arc arc, LM_PTR_state state, int* newvaluation, PTR_local_machine l,
		LM_list_of_variable * listvariables,LM_list_of_variable * currentvariable,AssignRet * globtoloc,AssignRet * gv,
		char * _action){
	
	LM_PTR_state new_state;
	LM_PTR_set_of_states sol= NULL;
	LM_PTR_set_of_states sol2 =NULL;

	if (currentvariable != NULL) {
		currentvariable->value = _TRUE_VALUE;
		sol = assign_succ(arc,state,newvaluation,l,listvariables,currentvariable->next,globtoloc,gv, _action);
		currentvariable->value = _FALSE_VALUE;
		sol2 = assign_succ(arc,state,newvaluation,l,listvariables,currentvariable->next,globtoloc,gv, _action);
		if (sol != NULL)
			LM_merge_set_of_states(sol,sol2);
		else
			sol = sol2;
	}
	else {
		new_state = construct_succ_assign_arc(arc,newvaluation,state,l,listvariables,globtoloc,gv, _action);

		if (new_state != NULL)
			LM_insert_state_into_set(&sol,new_state);	
	}
	return sol;
}

LM_PTR_set_of_states LM_post_assign_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l) {
        LM_PTR_set_of_states list_of_successor_states;
	int *newval =(int *) malloc(l->nLocVar * sizeof(int));
	int i;
	int val;
	AssignRet *list;
	AssignRet *copyassign = NULL;
	AssignRet *ptr;
	AssignRet *ptr2;
	AssignRet *gv;
	char *action= (char*)malloc(1024*sizeof(char));
	LM_list_of_variable *listvariables=NULL;
	
	action[0] = 0 ;

	/*copy of the local valuation*/
	for(i=0;i < l->nLocVar;i++) {
		newval[i] = state->val[i];
	}

	/*copy of the list of expressions*/
	(AssignRet *)ptr = state->cfNode->info;
	while (ptr != NULL) {
		ptr2 = (AssignRet *) malloc (sizeof(AssignRet));
		ptr2->varType = ptr->varType;
		ptr2->varId = ptr->varId;
		ptr2->exp = (ExpressionRet *)malloc(sizeof(ExpressionRet));
		ptr2->exp->type = ptr->exp->type;
		ptr2->exp->var = ptr->exp->var;
		ptr2->exp->not = ptr->exp->not;
		ptr2->next = copyassign;
		copyassign = ptr2;
		ptr = ptr->next;
	}

	/*treatment for the assignation from local variables to global variables*/
	
	if (control_flow_arc.action != NULL)/* We base the new action on the stuffs we have already done */
		strcpy(action,control_flow_arc.action);
	
	list = list_assign_local_to_global(&copyassign);
	while (list != NULL) {
		
		//if (strlen(action) != 0) /* If there's already something there */
		//	strcat(action, ", ") ;/*we add a comma */
			
		/* Now, for the assignation: first the variable name */
		/*strcat(action, tbsymbol_get_num(_glob_var_table,list->varId)->name);*/
		val = state->val[list->exp->var];
		if (list->exp->not == 1) {
			if (val == _TRUE_VALUE)
				val = _FALSE_VALUE;
			else
				val = _TRUE_VALUE;
		}
		
		/* Then, its value */
		
		if (val == _TRUE_VALUE)
			add_to_action(action, tbsymbol_get_num(_glob_var_table,list->varId)->name, _STRUE_) ;
		else
			add_to_action(action, tbsymbol_get_num(_glob_var_table,list->varId)->name, _SFALSE_) ;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations fom local variable to local variable*/
	list = list_assign_local_to_local(&copyassign);
	while (list != NULL) {
		val = state->val[list->exp->var];
		if (list->exp->not == 1) {
			if (val == _TRUE_VALUE)
				val = _FALSE_VALUE;
			else
				val = _TRUE_VALUE;
		}
		newval[list->varId] = val;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations from literals to local variables*/
	list = list_assign_literal_to_local(&copyassign);
	while (list != NULL) {
		if (list->exp->type == _LITTRUE)
			newval[list->varId] = _TRUE_VALUE ;	
		else
			newval[list->varId] = _FALSE_VALUE ;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations from global variable to global/local variable
	copyassign = list of assignation from global to local
	gv = list from global to global */

	/* We separate the list in two parts: gv contains the assignments from golbal to global. copyassign contains the rest:
	  the assignments from global variables to local variables */
	subList(&copyassign, &gv) ;
	/*creation of the list of global variable that must be tested*/
	ptr = gv;
	while (ptr != NULL) {
		LM_insert_variable(&listvariables, ptr->exp->var);
		ptr = ptr->next;
	}
	ptr = copyassign;
	while (ptr != NULL) {
		LM_insert_variable(&listvariables, ptr->exp->var);
	        ptr = ptr->next;
	}

	/*creation of the successors*/
	list_of_successor_states = assign_succ(control_flow_arc,state,newval,l,listvariables,listvariables,copyassign,gv, action);
	
	return list_of_successor_states;
}

//computes all the successors of an assignation node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_asign_node(LM_PTR_state state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL;
        LM_set_of_states * successor_states;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) {
		successor_states = LM_post_assign_node_arc(state,*control_flow_arc,l);
        	if (list_of_states != NULL)
                	LM_merge_set_of_states(list_of_states,successor_states);
		else
			list_of_states = successor_states;
                control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;
}


LM_PTR_state construct_succ_accept_arc(CF_arc arc, LM_PTR_state state,PTR_local_machine l,
		                LM_list_of_variable * listvariables){

        LM_PTR_trans trans;
        LM_PTR_state successor_state;
        LM_list_of_variable * ptr;
        char *msg = (char *)malloc(sizeof(char) * 500);
        int i;
	int *newval;
	
	//copy of the local valuation
	newval = (int *)malloc(sizeof(int) * l->nLocVar);
	for(i=0;i < l->nLocVar;i++) {
		newval[i] = state->val[i];
        }
	
        //construction of the message 
        strcpy(msg, arc.msg);
        ptr = listvariables;
	while (ptr != NULL) {
	        if (ptr->value == _TRUE_VALUE) 
	        	strcat(msg,"T");
		
		else 
		        strcat(msg,"F");
		ptr = ptr->next;
        }
		
	strcat(msg, "?") ;
			
	//for assignation from accept to local
	ptr = listvariables;
	while (ptr != NULL) {
	       	newval[ptr->varId] = ptr->value;
	       	ptr = ptr->next;
	}
	//construction of the successor
	successor_state = LM_get_state(l,arc.extr,newval);
	if (successor_state == NULL) {
	        successor_state = LM_create_state(arc.extr,newval,l->nLocVar);
	        LM_add_state (l, successor_state);
		trans = LM_add_trans (state, successor_state,arc.action,arc.guard, msg);
	}
	else {
	      trans = LM_add_trans (state, successor_state,arc.action,arc.guard, msg);
	      successor_state = NULL;
	}
	return successor_state;
}
	                                                                                                                                                                                                                                                                                                                                                                                                                                                        


LM_PTR_set_of_states accept_succ(CF_arc arc, LM_PTR_state state, PTR_local_machine l,
		                LM_list_of_variable * listvariables,LM_list_of_variable * currentvariable){

        LM_PTR_state new_state;
        LM_PTR_set_of_states sol= NULL;
        LM_PTR_set_of_states sol2 =NULL;

        if (currentvariable != NULL) {
                currentvariable->value = _TRUE_VALUE;
                sol = accept_succ(arc,state,l,listvariables,currentvariable->next);
                currentvariable->value = _FALSE_VALUE;
                sol2 = accept_succ(arc,state,l,listvariables,currentvariable->next);
                if (sol != NULL)
                        LM_merge_set_of_states(sol,sol2);
                else
                        sol = sol2;
        }
        else {
                new_state = construct_succ_accept_arc(arc,state,l,listvariables);

                if (new_state != NULL)
                        LM_insert_state_into_set(&sol,new_state);
	        }
	        return sol;
}


LM_PTR_set_of_states LM_post_accept_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l) {
	VarRet *ptr;
	LM_PTR_set_of_states sol;
	LM_list_of_variable *listvariables=NULL;

	ptr = (VarRet *)state->cfNode->info;
	while(ptr != NULL) {
		LM_insert_variable(&listvariables,ptr->varId);
		ptr = ptr->next;
	}
	sol = accept_succ(control_flow_arc,state,l,listvariables,listvariables);
	return sol;
}

//computes all the successors of an accept node
//returns only the successors that are new nodes
LM_set_of_states * LM_post_accept_node(LM_PTR_state state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL;
        LM_set_of_states * successor_states;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) {
                successor_states = LM_post_accept_node_arc(state,*control_flow_arc,l);
                if (list_of_states != NULL)
	                LM_merge_set_of_states(list_of_states,successor_states);
                else
                        list_of_states = successor_states;
                control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;
}

/*This function is called on a CFG arc of a choice statement, when we assume that the originating state of the arc satisfies the
condition of this branch of the choice. It returns the set of LM states that are reached by this branch, if they didn't exist in the
LM yet*/
LM_PTR_set_of_states LM_post_choice_assign(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l)
{
        LM_PTR_set_of_states list_of_successor_states;
	int *newval =(int *) malloc(l->nLocVar * sizeof(int));
	int i;
	int val;
	AssignRet *list;
	AssignRet *copyassign = NULL;
	AssignRet *ptr;
	AssignRet *ptr2;
	AssignRet *gv;
	char *action= (char*)malloc(1024*sizeof(char));
	LM_list_of_variable *listvariables=NULL;
	
	action[0] = 0 ;

	/*copy of the local valuation*/
	for(i=0;i < l->nLocVar;i++) {
		newval[i] = state->val[i];
	}

	/*copy of the list of expressions for the assignation. This is to be found in the arc too (unlike the assignation)*/
	(AssignRet *)ptr = ((GuardedAssignInfo *)control_flow_arc.info)->assign;
	while (ptr != NULL) {
		ptr2 = (AssignRet *) malloc (sizeof(AssignRet));
		ptr2->varType = ptr->varType;
		ptr2->varId = ptr->varId;
		ptr2->exp = (ExpressionRet *)malloc(sizeof(ExpressionRet));
		ptr2->exp->type = ptr->exp->type;
		ptr2->exp->var = ptr->exp->var;
		ptr2->exp->not = ptr->exp->not;
		ptr2->next = copyassign;
		copyassign = ptr2;
		ptr = ptr->next;
	}

	/*treatment for the assignation from local variables to global variables*/
	
	if (control_flow_arc.action != NULL)/* We base the new action on the stuffs we have already done */
		strcpy(action,control_flow_arc.action);
	
	list = list_assign_local_to_global(&copyassign);
	while (list != NULL) {
		
		/* Now, for the assignation: first the variable name */
		/*strcat(action, tbsymbol_get_num(_glob_var_table,list->varId)->name);*/
		val = state->val[list->exp->var];
		if (list->exp->not == 1) {
			if (val == _TRUE_VALUE)
				val = _FALSE_VALUE;
			else
				val = _TRUE_VALUE;
		}
		
		/* Then, its value */
		
		if (val == _TRUE_VALUE)
			add_to_action(action, tbsymbol_get_num(_glob_var_table,list->varId)->name, _STRUE_) ;
		else
			add_to_action(action, tbsymbol_get_num(_glob_var_table,list->varId)->name, _SFALSE_) ;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations fom local variable to local variable*/
	list = list_assign_local_to_local(&copyassign);
	while (list != NULL) {
		val = state->val[list->exp->var];
		if (list->exp->not == 1) {
			if (val == _TRUE_VALUE)
				val = _FALSE_VALUE;
			else
				val = _TRUE_VALUE;
		}
		newval[list->varId] = val;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations from literals to local variables*/
	list = list_assign_literal_to_local(&copyassign);
	while (list != NULL) {
		if (list->exp->type == _LITTRUE)
			newval[list->varId] = _TRUE_VALUE ;	
		else
			newval[list->varId] = _FALSE_VALUE ;
		ptr = list->next;
		free(list->exp);
		free(list);
		list = ptr;
	}

	/*treatment for the assignations from global variable to global/local variable
	copyassign = list of assignation from global to local
	gv = list from global to global */

	/* We separate the list in two parts: gv contains the assignments from golbal to global. copyassign contains the rest:
	  the assignments from global variables to local variables */
	subList(&copyassign, &gv) ;
	/*creation of the list of global variable that must be tested*/
	ptr = gv;
	while (ptr != NULL) {
		LM_insert_variable(&listvariables, ptr->exp->var);
		ptr = ptr->next;
	}
	ptr = copyassign;
	while (ptr != NULL) {
		LM_insert_variable(&listvariables, ptr->exp->var);
	        ptr = ptr->next;
	}

	/*creation of the successors*/
	list_of_successor_states = assign_succ(control_flow_arc,state,newval,l,listvariables,listvariables,copyassign,gv, action);
	
	return list_of_successor_states;

}


/* This is the post function applied to each arc going out of a CHOICE node */
LM_PTR_set_of_states LM_post_choice_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l) {
        LM_PTR_set_of_states successor_states;
	int *test;

	if (((GuardedAssignInfo*)control_flow_arc.info)->test == NULL) {/* No condition on the arc */
		successor_states = LM_post_choice_assign(state,control_flow_arc,l);
	}
	else {	/* there is a condition, we have to evaluate it */
			
		test = transform_expression(((GuardedAssignInfo *)control_flow_arc.info)->test, l->nLocVar);		
		if (test_value(state->val,test,l->nLocVar) == 1) { /* If the local values in this state correspond to the
		valuation requested by the condition in the choice, we must compute the state resulting of the assignation */
		
			successor_states = LM_post_choice_assign(state,control_flow_arc,l);
			
/*			successor_state = LM_get_state (l,control_flow_arc.extr,state->val);
*	        	if (successor_state == NULL) {// If this state doesn't exist in the LM yet, we have to create it 
*       	        	successor_state = LM_create_state(control_flow_arc.extr,state->val,l->nLocVar);
*	                	LM_add_state (l, successor_state);
*		                trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
*		        }
*        		else {// Otherwise, we simply create the transition 
*                		trans = LM_add_trans (state, successor_state, control_flow_arc.action, control_flow_arc.guard, control_flow_arc.msg);
*		                successor_state = NULL;
*		        }
*/
		}
		else 
			successor_states = NULL;
		free(test);
	}
        return successor_states;

}

/* This is the post function one applies to a CHOICE node */
LM_set_of_states * LM_post_choice_node(LM_PTR_state state, PTR_local_machine l) {
        CF_PTR_arc control_flow_arc;
        LM_set_of_states * list_of_states = NULL;
        LM_set_of_states * successor_states;

        control_flow_arc = state->cfNode->f_arc;
        while (control_flow_arc != NULL) {
                successor_states = LM_post_choice_node_arc(state,*control_flow_arc,l);
                if (list_of_states != NULL)
	                LM_merge_set_of_states(list_of_states,successor_states);
                else
                        list_of_states = successor_states;
                control_flow_arc = control_flow_arc->next;
        }
        return list_of_states;

}


//compute the set of successors of a state
//returns only the new states
LM_set_of_states * LM_post_node(LM_state * state, PTR_local_machine l) {
	LM_set_of_states * list_of_states = NULL;

	switch (state->cfNode->mark) {
		case _NORMAL_NODE :
			list_of_states = LM_post_normal_node(state,l);
			break;
		case _IF_NODE :
			list_of_states = LM_post_if_node(state,l);
			break;
		case _WHILE_NODE : 
			list_of_states = LM_post_while_node(state,l);
			break;
		case _RDV_NODE :
			list_of_states = LM_post_rdv_node(state,l);
			break ;
		case _ASSIGN_NODE :
			list_of_states = LM_post_asign_node(state,l);
			break;
		case _ACCEPT_NODE : 
			list_of_states = LM_post_accept_node(state,l);
			break ;
		case _CHOICE_NODE :
			list_of_states = LM_post_choice_node(state, l) ;
	}
	return list_of_states;
}

//compute the successors states of the states in list
//returns only the new states
LM_set_of_states * LM_post(LM_set_of_states * list,PTR_local_machine l) {
	LM_set_of_states * successors = NULL;
	LM_set_of_states * successors_states;
	LM_set_of_states *ptr;

	ptr = list;
	while (ptr != NULL) {
		successors_states = LM_post_node(ptr->state,l);
		if (successors_states != NULL) {
			if (successors == NULL)
				successors = successors_states;
			else
				LM_merge_set_of_states(successors,successors_states);
		}
		ptr = ptr->next;
	}
	return successors;
}

void LM_insert_variable(LM_list_of_variable ** list, int varId) {
	LM_list_of_variable *newelement;
	LM_list_of_variable *ptr = *list;
	int stillexists = 0;
	
	while ((ptr != NULL) && (stillexists == 0)) {
		if (ptr->varId == varId) 
			stillexists = 1;
		else
			ptr = ptr->next;
	}
	if (stillexists == 0) { /* We have to insert this element in the tail of the list to 
				respect the ordering of the original list */
		newelement = (LM_list_of_variable *) malloc(sizeof(LM_list_of_variable));
		newelement->varId = varId;
		newelement->next = NULL;
		newelement->value = _NO_VALUE;
		ptr = *list ;
		if (ptr == NULL)
		{
			*list = newelement ;
		}
		else
		{
			while (ptr->next != NULL)
				ptr = ptr->next ;
			
			ptr->next = newelement ;
		}
	}
}

int value_of_var(LM_list_of_variable * listvariable, int varId) {
	LM_list_of_variable * ptr =listvariable;
	int sol = _NO_VALUE;
	
	while (ptr != NULL) {
		if (ptr->varId == varId) {
			sol = ptr->value;
			ptr = NULL;
		}
		else 
			ptr = ptr->next;
	}
	return sol;
}
