/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _GLOBALMACHINE_H_
#define _GLOBALMACHINE_H_
#include "controlflow.h"
#include "tbsymbol.h"
#include "stringmanip.h"

#define _COMMON_ACTIONS_ "if, elseif, endif, while, endwhile, goto, skip, lock, sleep, unlock, sleep, assign, end_assign"

struct s_LM_trans ;


/* One local machine state */
typedef struct s_LM_state {
	CF_PTR_node cfNode ; /* The point of code */
	unsigned int * val ; /* The valuation of the local variables */
	struct s_LM_trans * firstTrans ; /* The first transition of the list of outgoing transitions */
	struct s_LM_state * next ;
	} LM_state ;
	
typedef LM_state * LM_PTR_state ;

/* One transition of a local machine */
typedef struct  s_LM_trans {
	LM_PTR_state extr ; /* The destination state */
	char * action ;
	char * guard ;
	char * label ;
	struct s_LM_trans * next ;
	} LM_trans ;
	
typedef LM_trans * LM_PTR_trans  ;
	

/* Structures used to store the list of nodes in a LM
  These structures allow a more efficient search, because
  all the nodes in the sublists share the same cfg node.
 */

typedef struct s_LM_state_list {
	CF_PTR_node cfNode ; /* The common cfg node */
	LM_PTR_state first ; /* The first state of this sublist */
	struct s_LM_state_list * next ;
	} LM_state_list ;

typedef LM_state_list * LM_PTR_state_list ;

/* One local machine */
typedef struct s_local_machine {
	LM_PTR_state_list list ; /* The list of nodes of the LM */
	int nLocVar ; /* Number of local variables in this LM */
	char * name ; /* actually, the name of the thread */
	struct s_local_machine * next ; /* To chain the LM in the list of a GM */
	LM_PTR_state init_state ; /* This field will be set for the main thread only */
	} local_machine ;

typedef local_machine * PTR_local_machine;

/* One global machine */
typedef struct s_global_machine {
	PTR_local_machine first ; /* The head of the LM list */
	int nGlobVar ; /* The number of global variables */ 
	} global_machine ;

typedef global_machine * PTR_global_machine ;

/*set of states*/
typedef struct LM_set_of_states {
	LM_PTR_state state;
	struct LM_set_of_states * next;
} LM_set_of_states;
	
typedef LM_set_of_states * LM_PTR_set_of_states ;

/*structure to maintain the initial local states for a thread*/
typedef struct LM_initial_states {
	int nb_variable;
        char * thread; //name of the thread
        LM_PTR_state first_state;  //first initial local state
        struct LM_initial_states * next; 
} LM_initial_states;

typedef LM_initial_states * PTR_LM_initial_states;

typedef struct LM_list_initial_states {
	PTR_LM_initial_states first_element;
} LM_list_initial_states;


typedef LM_list_initial_states * PTR_LM_list_initial_states;


//structure for the assignation
typedef struct LM_list_of_variable {
	int varId;
	int value;
	struct LM_list_of_variable * next;
} LM_list_of_variable;

/*The functions for the local machines */

//some global variable used to store the initial local states of the local machines
PTR_LM_list_initial_states list_of_initial_states;



PTR_local_machine  LM_create_machine (int n) ;
LM_PTR_state LM_create_state (CF_PTR_node n, unsigned * v, int nVar) ;
void LM_add_state (PTR_local_machine l, LM_PTR_state s) ;
LM_PTR_state LM_get_state (PTR_local_machine l, CF_PTR_node n, unsigned * v) ;
LM_PTR_trans LM_add_trans (LM_PTR_state o, LM_PTR_state e, char * action, char * guard, char * label) ;

PTR_LM_initial_states LM_get_local_states(LM_list_initial_states list, char *local_machine);
PTR_LM_initial_states LM_create_initial_states(char * name);
void LM_insert_local_state(LM_PTR_state state, PTR_LM_initial_states list);
int LM_contain_state(LM_PTR_state state, PTR_LM_initial_states list);

void LM_insert_list_local_states(PTR_LM_list_initial_states list, char *local_machine);
void create_list_local_states(PTR_LM_list_initial_states *list);

int LM_same_local_state(LM_state state1, LM_state state2, int nb_value);


void LM_insert_state_into_set(LM_set_of_states **set,LM_PTR_state state);
void LM_merge_set_of_states(LM_set_of_states *set1, LM_set_of_states *set2);

PTR_global_machine GM_generate(void) ;
LM_PTR_state LM_post_normal_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l);
LM_set_of_states * LM_post_normal_node(LM_PTR_state state, PTR_local_machine l);
LM_PTR_state LM_post_if_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l);
LM_set_of_states * LM_post_if_node(LM_PTR_state state, PTR_local_machine l);
LM_PTR_state LM_post_while_node_arc(LM_PTR_state state, CF_arc control_flow_arc, PTR_local_machine l);
LM_set_of_states * LM_post_while_node(LM_PTR_state state, PTR_local_machine l);

LM_set_of_states * LM_post_node(LM_PTR_state  state, PTR_local_machine l);
LM_set_of_states * LM_post(LM_set_of_states * list,PTR_local_machine l);

void LM_insert_variable(LM_list_of_variable ** list, int varId);
int value_of_var(LM_list_of_variable * listvariable, int varId);

char * LM_get_state_name(LM_PTR_state s, PTR_local_machine L) ;
void LM_write_to_file(FILE * f, PTR_local_machine L) ;
void GM_write_to_file(PTR_global_machine G, FILE * f, char * name) ;

void LM_generate_rdv_msg(char * msg, int n) ;
#endif
