/*  Copyrigth 2003 Laurent Van Begin and Gilles Geeraerts
 *
 *  This file is part of the CBP2GM compiler.
 *
 *  The CBP2GM compiler is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The CBP2GM compiler is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CBP2GM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "CBP.h"

/* This list will (hopefully) contain all the messages that need to appear in the 'Action' section
   of the global machine file. */
PTR_msgList mList= NULL ;

/* Receives a list of arcs bearing each a guarded assignation, and returns the sub-CFG corresponding to it */
sub_graph __guarded_assign(GuardedAssignList * list) 
{
	GuardedAssignList * p = list ;
	CF_PTR_node dummy ;
	sub_graph * sg =(sub_graph *)malloc(sizeof(sub_graph)) ;
	
	/* We create the intial node and the dummy node */
	sg->begin = CF_create_node(-1, NULL, _CHOICE_NODE) ;
	dummy = CF_create_node(-1, NULL, _NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
	CF_insert_node(cfg, dummy) ;
	
	/* we create the output arc and connect it to the dummy node */
	sg->out = CF_create_arc(NULL, "", "", "endchoice") ;
	CF_set_arc(dummy, sg->out) ;
	
	while (p != NULL)
	{
		CF_set_arc(sg->begin, p->arc) ;
		CF_set_extr(p->arc, dummy) ;
		p = p->next ;
	}
	
	return *sg ;
}

/* This function creates one arc for one guarded asignation and returns it into the GuardedAssignList structure.
The info field of the arc possibly points to a condition on local variables 
The returned structure also contains a list of assignations that remain to be treated (this information will be put into
the info field of the node). */
GuardedAssignList * ga_build_arc(ExpressionRet * exp, AssignRet * assign)
{
	char * cond ;
	T_PTR_tbsymbol_entry entry;
	GuardedAssignList * elem = (GuardedAssignList *)malloc(sizeof(GuardedAssignList)) ;
	AssignRet * lit;
	AssignRet * p ;
	GuardedAssignInfo * info = (GuardedAssignInfo *)malloc(sizeof(GuardedAssignInfo)) ;
	char * action = (char *)malloc(500*sizeof(char)) ; // where we will stock the action

	info->assign = NULL ;
	info->test = NULL ;
	action[0] = 0 ;
	cond = "" ;
	
	elem->arc = CF_create_arc(NULL, "", "", "choice") ;
	CF_set_info_arc(elem->arc, info) ;
	
	/* First of all, let's look at the expression*/
	switch (exp->type) 
	{
		case _LITFALSE: /* This branch can never be executed -> we don't create anything*/
			free(elem) ;
			return NULL ;
			break ;
		case _GLOBVAR: /* we must test it on the arc */
			/* Let's get its name from the symbol table*/
			entry = tbsymbol_get_num( _glob_var_table,exp->var);
			if (entry == NULL)
				err_sys("ga_build_arc :: variable does not exist !");
			/* We can now put the expression on the arc */
			cond = (char*)malloc(sizeof(char)*(sizeof(entry->name)+_NOTSIZE_+1)) ; 
			cond[0] = 0 ;
			add_to_guard(cond, entry->name, exp->not) ;
			break ;
		case _LOCVAR: /* This will be handeled later -> we retain it as an info on the arc */
			info->test = malloc(sizeof(ExpressionRet)) ;
			memcpy(info->test, exp, sizeof(ExpressionRet)) ;
			break ;
		case _LITTRUE: /* Nothing to do */
			break ;
		default:
			err_sys("ga_build_arc :: internal error: unknown type of expression !") ;
	}
	
	/*Now, we take care of the assignation. We identify
	the set of assignments from literals to global variables.
	The rest of the list has to be kept in the info field of the first node of the 
	subgraph */
	list_assignation_literals_to_global(&assign, &lit) ;

	p = lit ;
	
	while (p)
	{
		/*Let's first fetch the variable name into the symbol table*/
		entry = tbsymbol_get_num(_glob_var_table, p->varId) ;	

		/*As well as the value to assign*/
		if (p->exp->type == _LITTRUE)
			add_to_action(action, entry->name, _STRUE_) ;
		else if (p->exp->type == _LITFALSE)
			add_to_action(action, entry->name, _SFALSE_) ;
		else
			err_sys("Internal error: __assign : I found a non-litteral in the 'lit' list") ;

		p = p->next ;
	}
	
	/* cond and action respectively contain the condition and action computed so far. We put them on the arcs */
	CF_set_label_arc(elem->arc, cond, action, "choice") ;
	
	/* Now assign contains the remaining assignations*/
	info->assign = assign ;

	return elem ;
}

sub_graph __if(ExpressionRet * expression, sub_graph *ifbranch, sub_graph *elsebranch)
{
	CF_PTR_arc ifbranch_arc;
	CF_PTR_arc elsebranch_arc ;
	CF_PTR_arc output ;
	T_PTR_tbsymbol_entry entry;
	char *cond;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;
	void * info;
	CF_PTR_node dummy ;
	
	/* We first have to create an input node and a dummy node*/
	sg->begin = CF_create_node(-1, NULL, _IF_NODE) ;
	dummy = CF_create_node(-1, NULL, _NORMAL_NODE) ;
	CF_insert_node(cfg, dummy) ;
	
	/* Now, we put an arc from sg->begin to the subgraphs for the if and else branches 
	   'unless the epression forbids it) */
	if(expression->type != _LITFALSE) /* we can make an if arc */
	{
		ifbranch_arc = CF_create_arc(ifbranch->begin, "", "", "if") ;
		CF_set_arc(sg->begin, ifbranch_arc) ;
	}
	if(elsebranch) /* If there was an else specified */
	{
		if(expression->type != _LITTRUE) /* we can make an else arc */
		elsebranch_arc = CF_create_arc(elsebranch->begin, "", "", "else") ;
		CF_set_arc(sg->begin, elsebranch_arc) ;
	}
	else /* We direct it to the dummy node*/
	{
		elsebranch_arc = CF_create_arc(dummy, "", "", "else") ;
		CF_set_arc(sg->begin, elsebranch_arc) ;
	}
	
	
	/* We connect the output arcs of the ifbranch and elsebranch subgraphs to the dummy node
	   (unless these arcs already exist) */
	
	if (!CF_get_extr(ifbranch->out))
		CF_set_extr(ifbranch->out, dummy) ;
	if (elsebranch && !CF_get_extr(elsebranch->out))
		CF_set_extr(elsebranch->out, dummy) ;

	/* And we put an 'ouput' arc going out of the dummy node */
	output = CF_create_arc(NULL, "", "", "endif") ;
	CF_set_arc(dummy, output) ;
	sg->out = output ;
	
	/* Now let's deal with the expression */

	if(expression->type == _LITFALSE || expression->type == _LITTRUE) /* we don't need to go any further */
		return _sg ;
	
	if (expression->type == _GLOBVAR) { /* Global variable -> we can handle the guard now */
		entry = tbsymbol_get_num( _glob_var_table,expression->var);
		if (entry == NULL)
			err_sys("__if :: variable does not exists !");
		/* We can now put the expression on the arcs */
		/* First on the if branch*/
		cond = (char*)malloc(sizeof(char)*(sizeof(entry->name)+_NOTSIZE_+1)) ; 
		cond[0] = 0 ;
		add_to_guard(cond, entry->name, expression->not) ;
		CF_set_label_arc(ifbranch_arc, cond, "", "if") ;
		/* then on the else */	
		cond[0] = 0 ;
		if (expression->not == 1)
			add_to_guard(cond, entry->name, 0) ;
		else
			add_to_guard(cond, entry->name, 1) ;
		CF_set_label_arc(elsebranch_arc, cond, "", "else") ;
	

	}
	/* The expression is star -> nothing to do (like true) */
	else if (expression->type == _STAR) {
		;
	}
	/* The expression is on local variable -> we will take care of it later */
	else if (expression->type == _LOCVAR) {
		/* We just put the info on the arcs */
		
		/* The if branch */
		info = malloc(sizeof(ExpressionRet)) ;
		memcpy(info, expression, sizeof(ExpressionRet)) ;
		CF_set_info_arc(ifbranch_arc,info);
		
		/* The else branch */
		info = malloc(sizeof(ExpressionRet)) ;
		memcpy(info, expression, sizeof(ExpressionRet)) ;
		if (expression->not == 1)
			((ExpressionRet*)info)->not = 0 ;
		else
			((ExpressionRet*)info)->not = 1 ;
		CF_set_info_arc(elsebranch_arc,info);

	}

	return _sg ;
}


	

sub_graph __while(ExpressionRet * expression, sub_graph _exp) {
	CF_PTR_arc exp_arc;
	CF_PTR_arc not_exp_arc;
	T_PTR_tbsymbol_entry entry;
	char *cond;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;
	sub_graph * exp = &_exp ;

	//construction of the input node
	sg->begin = CF_create_node(-1,NULL,_NORMAL_NODE);
	
	CF_insert_node(cfg, sg->begin) ;
	
	// If the expression is false -> we can directly skip to the end of the while
	if (expression->type == _LITFALSE)
	{
		exp_arc = CF_create_arc(NULL, "", "", "endwhile") ;
		CF_set_arc (sg->begin, exp_arc) ;
		sg->out=exp_arc ;
		return _sg ;
	}
	
	//construction of the arc to the instructions
	exp_arc = CF_create_arc(exp->begin,"","","while");
	CF_set_arc(sg->begin,exp_arc);
	
	// The output arc of the instructions must have its extremity redirected to the input node of the while
	// unless it has already an extremity (it is the case if the last instruction is a goto)
	if(!CF_get_extr(exp->out))
	{
		CF_set_extr(exp->out, sg->begin) ;
	}
	//CF_insert_node(cfg,node);
	
	//put the arc to exit the while
	if (expression->type != _LITTRUE) // If the expression is "true", we don't need the "false" branch
	{
		not_exp_arc = CF_create_arc(NULL, "","","endwhile");
		CF_set_arc(sg->begin,not_exp_arc);
		//the output arc will always be the not_exp_arc
		sg->out=not_exp_arc ;
	}
	else // otherwise, there is no output branch... just create a dummy one without origin
	{
		
		sg->out = CF_create_arc(NULL, "", "", "dummy") ;
		return _sg ;
	}

	
	//managing of the expression
	//if expression is a global variable
	if (expression->type == _GLOBVAR) {
		entry = tbsymbol_get_num( _glob_var_table,expression->var);
	        if (entry == NULL)
	        	err_sys("__if :: variable does not exists !");
		//put expression on the right arc
		if (expression->not == 0) {
			//label of the while arc
			cond = (char *)malloc(sizeof(char) * strlen(entry->name));
			strcpy(cond,entry->name);
			CF_set_label_arc(exp_arc,cond,"","while");
			free(cond);
			//label of the endwhile arc
			cond = (char *)malloc(sizeof(char) * (strlen(entry->name) + _NOTSIZE_+1));
			cond[0] = 0 ;
			add_to_guard(cond, entry->name, 1) ;
			CF_set_label_arc(not_exp_arc, cond, "", "endwhile");
			free(cond);
		}
		else {
			//label of the endwhile arc
                        cond = (char *)malloc(sizeof(char) * (strlen(entry->name) +_NOTSIZE_+1));
			cond[0] = 0 ;
			add_to_guard(cond, entry->name, 1) ;
                        CF_set_label_arc(exp_arc, cond, "", "while");
			free(cond);
			//label of the endwhile arc
			cond = (char *)malloc(sizeof(char) * strlen(entry->name));
			strcpy(cond,entry->name);
			CF_set_label_arc(not_exp_arc,cond,"","while");
			free(cond);
		}
	}
	else if (expression->type == _STAR) {
		//free(expression);
	}
	else if (expression->type == _LOCVAR) {
		sg->begin->info = malloc(sizeof(ExpressionRet)) ;
		memcpy(sg->begin->info, expression, sizeof(ExpressionRet)) ;
		sg->begin->mark = _WHILE_NODE;
	}
	return _sg ;	
}


sub_graph __skip(void)
{
	sub_graph sg ;
	sg.out = CF_create_arc(NULL,"", "", "skip" ) ;
	sg.begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg.begin) ;
	CF_set_arc(sg.begin, sg.out) ;
	return sg ;
	
}

sub_graph __goto( char * label)
{
	CF_PTR_node n ;
	int i = tbsymbol_get_index(_label_table, label) ;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;

	//let's find the extremity of the new arc
	if (i == -1) // node not found !!
	{	
		i = tbsymbol_insert_num(_label_table, label );
		n = CF_create_node(i, NULL,_NORMAL_NODE) ;
		CF_insert_node(cfg, n) ;
		//printf("__goto: not found node %s... creating a new one at position %i (adress: %x)\n", label, i, n) ;		
	}
	else
	{
		n = CF_get_node(cfg, i) ;
		//printf("__goto: found node %s...\n ", label) ;
	}
	//new node for the goto
	sg->out = CF_create_arc(n, "", "", "goto" ) ;
	
	sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_set_arc(sg->begin, sg->out) ;
	CF_insert_node(cfg, sg->begin) ;

	
	return _sg ;
}


sub_graph __lock(char *lock)
{
        char *guard;
        char *action;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;

	
	// no locks no chocolates
        if (lock == NULL)
                err_sys("__lock :: no lock locked !");
		
	//indeed...
	if (tbsymbol_select(_lock_table, lock) == NULL)
	{
		char * buf = (char *)malloc(1024*sizeof(char));
		sprintf(buf, "[%i] undeclared lock %s in unlock statement. Giving up.", lineno, lock) ;
		err_sys(buf) ;
	}
	
	
	//construction of the guard and the action
        guard = (char *)malloc(sizeof(char) *(strlen(lock)+_NOTSIZE_));
        action =(char *)malloc(sizeof(char) *(strlen(lock)+strlen("':=true")+1));
	guard[0] = 0 ;
	action[0] = 0 ;
	add_to_guard(guard, lock, 1) ;
	add_to_action(action, lock, _STRUE_) ;
	
	/*strcpy(guard,lock);
	strcpy(action,lock);
        strcat(action,"':=false");*/
	//creation of the output arc
        sg->out = CF_create_arc(NULL, guard, action, "lock" ) ;
	free(guard);
	free(action);
        sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
        CF_set_arc(sg->begin, sg->out) ;
	
	return _sg ;
}

sub_graph __unlock( char *lock)
{
        char *action;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;
	
	//no locks no chocolates
	if (lock == NULL)
                err_sys("__unlock :: no lock unlocked !");
	
	//indeed...
	if (tbsymbol_select(_lock_table, lock) == NULL)
	{
		char * buf = (char *)malloc(1024*sizeof(char));
		sprintf(buf, "[%i] undeclared lock %s in lock statement. Giving up.", lineno, lock) ;
		err_sys(buf) ;
	}

		
	//construction of the guard
	action =(char *)malloc(sizeof(char) *(strlen(lock)+strlen("':=false")+1));
	action[0] = 0 ;
	add_to_action(action, lock, _SFALSE_) ;
	/*strcpy(action,lock);
        strcat(action,"':=false");*/
	
	//creation of the output arc
	sg->out = CF_create_arc(NULL, "", action, "unlock" ) ;
        sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
        CF_set_arc(sg->begin, sg->out) ;
	
	return _sg ;
}

sub_graph __sleep( char *lock, char *msg)
{
	CF_PTR_node node_wait;
	CF_PTR_node node_lock;
	CF_PTR_arc arc;
	char *guard;
	char *action;
	char *sync;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;

	
	//not locks no chocolates
	if (lock == NULL)
                err_sys("__sleep :: no lock unlocked !");
	if (msg == NULL)
		err_sys("__sleep :: no message to be awaken !");
	//creation of the nodes corresponding to the sleep feature
	sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
	node_wait = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	node_lock = CF_create_node(-1,NULL,_NORMAL_NODE);
	
	//insertion of the nodes into the graph
	CF_insert_node(cfg,node_wait);
	CF_insert_node(cfg,node_lock);
	
	// arc from sleep to wait
	action = (char *)malloc(sizeof(char) * (strlen(lock)+strlen("':=false")+1));
	action[0] = 0 ;
	add_to_action(action, lock, _SFALSE_) ;

	/*strcpy(action,lock);
        strcat(action,"':=false");*/
	arc = CF_create_arc(node_wait,"",action,"sleep");
	free(action);
	CF_set_arc(sg->begin,arc);
	
	//arc for broadcast
	sync = (char *)malloc(sizeof(msg) + sizeof("??"));
	strcpy(sync,msg);
	strcat(sync,"??");
	arc = CF_create_arc(node_lock,"","", sync);
	free(sync);
	CF_set_arc(node_wait,arc);
	 
	
	//arc for asynchronous rendez-vous
        sync = (char *)malloc(sizeof(msg) + sizeof("^^"));
        strcpy(sync,msg);
        strcat(sync,"^^");
        arc = CF_create_arc(node_lock,"","", sync);
	free(sync);
        CF_set_arc(node_wait,arc);
	
	/* Don't forget to remember this message: we will need it when generating the global machines */
	msgList_insert(msg, &mList) ;

	//output arc
	action = (char *)malloc(strlen(lock)+strlen("':=true")+1);
	action[0] = 0 ;
	add_to_action(action, lock, _STRUE_) ;
        /*strcpy(action,lock);
        strcat(action,"':=false");*/
	guard = (char *)malloc(strlen(lock)+_NOTSIZE_+1);
	add_to_guard(guard, lock, 1) ;
	sg->out = CF_create_arc(NULL,guard, action, "lock" ) ; 
	free(guard);
	free(action);
        CF_set_arc(node_lock,sg->out);	
	
	return _sg ;
}


sub_graph __rdv( char * msg)
{

	char * buf = (char *)malloc(1024*sizeof(char)) ;

	sub_graph _sg ;
	sub_graph *sg = &_sg ;
	
	// The message should already be declared
	
	if (tbsymbol_select(_msg_table, msg) == NULL)
	{
		sprintf(buf, "[%i] Undeclared message %s in rendezvous statement. Giving up", lineno, msg) ;
		err_sys(buf) ;
	}

	//We need to create one node with one outgoing arc
	//The msg on the arc will be label!
	
	//we create the label

	
	strcpy(buf,msg);
	strcat(buf,"!") ;
	
	/* Don't forget to remember this message: we will need it when generating the global machines */
	msgList_insert(msg, &mList) ;

	
	//then the arc
	sg->out = CF_create_arc(NULL, "", "", buf) ;
	free(buf);
	// and the node
	sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
	
	// finally, we connect the arc to the node
	CF_set_arc(sg->begin, sg->out) ;
	return _sg ;
}

sub_graph __start(char * msg)
{
	char * newmsg ;
	//The message should be a known thread name


	if (tbsymbol_select(_thread_table, msg)==NULL)
	{
		char * buf = (char*)malloc(1024*sizeof(char)) ;
		sprintf(buf, "[%i] Unknown thread name %s in start statement. Giving up !", lineno, msg) ;
		err_sys(buf) ;
	}

	// All right, then we have to make a rendez-vous on _start_msg
	
	newmsg = (char *)malloc(sizeof(msg)+sizeof("_start_")+sizeof(char)) ;
	strcpy(newmsg, "_start_") ;
	strcat(newmsg, msg) ;

	//Now let's add this new message to the message table
	
	tbsymbol_insert(_msg_table, newmsg) ;

	// After all, a start is just a rendez-vous !!
	
	return __rdv(newmsg) ;
}

sub_graph __wakeup(char * msg)
{
        //We need to create one node with one outgoing arc
        //The msg on the arc will be label!

	//we create the label
        char * buf=(char *)malloc(sizeof(char)*(strlen(msg)+2)) ;
	sub_graph _sg ;
	sub_graph *sg = &_sg ;

	
	strcpy(buf,msg);
	strcat(buf,"!") ;

	/* Don't forget to remember this message: we will need it when generating the global machines */
	msgList_insert(msg, &mList) ;

        //then the arc
        sg->out = CF_create_arc(NULL, "", "", buf) ;
	free(buf);
        // and the node
        sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;

        // finally, we connect the arc to the node
        CF_set_arc(sg->begin, sg->out) ;
	
	return _sg ;
}


sub_graph __wakeupall(char * msg)
{
	//We need to create one node with one outgoing arc
	//The msg on the arc will be label!!
	
	sub_graph _sg ;
	sub_graph *sg = &_sg ;

	
	//we create the label
	char * buf=(char *)malloc(sizeof(char)*(strlen(msg)+3)) ;
	strcpy(buf,msg);
	strcat(buf,"!!") ;
	
	/* Don't forget to remember this message: we will need it when generating the global machines */
	msgList_insert(msg, &mList) ;
	
	//then the arc
	sg->out = CF_create_arc(NULL, "", "", buf) ;
	free(buf);
	// and the node
	sg->begin = CF_create_node(-1, NULL,_NORMAL_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
	
	// finally, we connect the arc to the node
	CF_set_arc(sg->begin, sg->out) ;
	
	return _sg ;
}


sub_graph __accept( char * msg)
{
	//We need to create one node with one outgoing arc
	//The msg on the arc will be label?
	
	sub_graph _sg ;
	sub_graph *sg = &_sg ;
	
	//we create the label
	char * buf=(char *)malloc(sizeof(char)*(strlen(msg)+2)) ;
	strcpy(buf,msg);
	/*strcat(buf,"?") ;*/

	/* Don't forget to remember this message: we will need it when generating the global machines */
	msgList_insert(msg, &mList) ;
	
	//then the arc
	sg->out = CF_create_arc(NULL, "", "", buf) ;
	free(buf);
	// and the node
	sg->begin = CF_create_node(-1, NULL,_ACCEPT_NODE) ;
	CF_insert_node(cfg, sg->begin) ;
	
	// finally, we connect the arc to the node
	CF_set_arc(sg->begin, sg->out) ;
	
	return _sg;
}

sub_graph __rdv_param(char * msg, ExpListRet * param)
{
	/*
	param contains a list representing all the expression we want to pass
	through the rendez-vous. 
	To generate the message that will be actually issued, we need to know
	the values of the local variables. Therefore, we keep the list of expression
	in the node/arc (??) representing the rendez-vous for subsequent use (the messages
	will be constructed once we generate the global machine).
	*/

	char * buf;
	sub_graph _sg;
	sub_graph * sg= & _sg;

	if (tbsymbol_select(_msg_table, msg) == NULL)
	{
		buf = (char *)malloc(1024*sizeof(char)) ;
		sprintf(buf, "[%i] Undeclared message %s in rendezvous statement. Giving up", lineno, msg) ;
		err_sys(buf) ;
	}
	//we create the node
	sg->begin = CF_create_node(-1, param, _RDV_NODE);
	CF_insert_node(cfg,sg->begin);
	// and the arc
	sg->out = CF_create_arc(NULL, "", "", msg) ;
	
	free(msg);

	//finally, we connect the arc to the node
	CF_set_arc(sg->begin, sg->out) ;

	return _sg ; 
}

sub_graph __start_param(char * msg, char * exp)
{
	char * newmsg  ;
	//has the thread name been declared ?
	if (tbsymbol_select(_thread_table, msg)==NULL)
	{
		char * buf = (char*)malloc(1024*sizeof(char)) ;
		sprintf(buf, "[%i] Unknown thread name %s in start statement. Giving up !", lineno, msg) ;
		err_sys(buf) ;
	}

	//Let's forge the message we will use in the rendez-vous

	newmsg = (char *)malloc(sizeof(msg)+sizeof(exp)+sizeof("_start_")+sizeof(char)) ;
	strcpy(newmsg, "_start_") ;
	strcat(newmsg, msg) ;
	strcat(newmsg, exp );
	
	//Now let's add this new message to the message table

	tbsymbol_insert(_msg_table, newmsg) ;

	return __rdv(newmsg) ;
}

sub_graph __accept_param(char * msg, VarRet * vars)
{
	/* We receive, in var, the list of local variables that are going to store
	 the values transmitted through the rendez-vous
	 These variables are identified thanks to their symbol table id
	 We should create as many synchronization messages as there can be values
	 for this set of local variables
	 To avoid this, we just attach the list of variables to the node
	 in order to construct the requested set of messages later.*/
	sub_graph g ;
	g = __accept(msg) ;
	
	/*Now we have a sub-graph with "msg" on the arc, but it will be changed later
	Let's add the list of variables to the info field of the incoming node of this 
	sub-graph, and mark the node. */
	
	g.begin->info = vars ;
	g.begin->mark = _ACCEPT_NODE ;
	return g ;
}

void list_assignation_literals_to_global(AssignRet ** assign, AssignRet ** lit) {
	AssignRet *ptr;
	AssignRet *next;
	AssignRet *rest = NULL;

	ptr = *assign;
	*lit = NULL;
	while(ptr != NULL) {
		next = ptr->next;
		if ((ptr->varType == _GLOBVAR) && ((ptr->exp->type == _LITTRUE) || (ptr->exp->type == _LITFALSE))) {
			ptr->next = *lit;
			*lit = ptr;
		}
		else {
			ptr->next = rest;
			rest = ptr;
		}
		ptr = next;
	}
	*assign = rest;
}

sub_graph __assign(AssignRet * assign) 
{
	/*We receive a list containing pairs (ident, Expression). We identify
	the set of assignments literals to global variables.
	The rest of the list is kept in the info field of the first node of the 
	subgraph */
	
	/*First: isolate the sub-list involving only global variables*/
	AssignRet * lit;
	AssignRet * p ;
	T_PTR_tbsymbol_entry tbe ;
	char * action = (char *)malloc(500*sizeof(char)) ; // where we will stock the action
	sub_graph sg ;
	action[0] = 0 ;

	//management of the assignation from literals to global variables
	list_assignation_literals_to_global(&assign, &lit) ;

	/*Then, generate the arcs*/
	
	p = lit ;
	
	while (p)
	{
		/*Let's first fetch the variable name into the symbol table*/
		tbe = tbsymbol_get_num(_glob_var_table, p->varId) ;	

		/*As well as the value to assign*/
		if (p->exp->type == _LITTRUE)
			add_to_action(action, tbe->name, _STRUE_) ;
		else if (p->exp->type == _LITFALSE)
			add_to_action(action, tbe->name, _SFALSE_) ;
		else
			err_sys("Internal error: __assign : I found a non-litteral in the 'lit' list") ;

		p = p->next ;
	}
	
	/*Let's now create the input node of the subgraph for this assignment*/
	
	sg.begin = CF_create_node(-1, assign, _ASSIGN_NODE) ;
	CF_insert_node(cfg, sg.begin) ;
	
	/*Now, let's just add an arc going out of sg.begin, and we are done*/
	sg.out = CF_create_arc(NULL, "", action, "assign") ;
	CF_insert_arc(cfg, sg.begin, sg.out) ;

	return sg ;
}

/*
* Recursive function used to build the set of arcs for an assignment glob->glob
e.g.:

A=!B and C= D ===>

        /---- B, C, A:= false, D:=true ------------>\
       /----- B, not(C), A:= false, D:=false ------->\
(n)--->\----- not(B), C, A:= true, D:= true -------->/--->(e)--->
        \---- not(B), not(C) A:= true, D:= false -->/
    
    The list of assignment is contained in l
    buf_a and buf_g are used to store intermediate values for the guards and assign 
*/
void rec_build_arcs(CF_PTR_node n, CF_PTR_node e, AssignRet * l, char * buf_a, char * buf_g)
{
	T_PTR_tbsymbol_entry tbsource ;
	T_PTR_tbsymbol_entry tbdest ;
	char * work_buf_a = (char *)malloc(500*sizeof(char)) ;
	char * work_buf_g = (char *)malloc(500*sizeof(char)) ;
	CF_PTR_arc a ;

	if (!l) /*If the list is empty, we just generate an arc from n to e with the computed info*/
	{
		a = CF_create_arc(e, buf_g, buf_a, "");
		CF_insert_arc(cfg, n, a) ;
	}
	else /*Otherwise, we go recursive...*/
	{
	
		/*Let's fetch the source and dest variables*/
		tbdest = tbsymbol_get_num(_glob_var_table, l->varId) ;
		tbsource = tbsymbol_get_num(_glob_var_table, l->exp->var) ;
			
		/*We need a working copy of the buffers*/
		
		strcpy(work_buf_a, buf_a) ;
		strcpy(work_buf_g, buf_g) ;	
				
		/*We have to create the two cases for the assignment in l*/
		
		/*step  1: the source variable is true*/
		
		add_to_guard(work_buf_g, tbsource->name, 0)  ;
		
		/*Let's analyse the expression to see what to assign*/
		if (l->exp->not == 0) /*dest = source*/
		{	
			add_to_action(work_buf_a, tbdest->name, _STRUE_) ;
/*			strcat(work_buf_a, tbdest->name) ;
			strcat(work_buf_a, " := true") ;*/
		}
		else /*dest = not(source)*/
		{			
			add_to_action(work_buf_a, tbdest->name, _SFALSE_) ;
/*
			if(strlen(work_buf_a) != 0)
				strcat(work_buf_a, ", ") ;
			strcat(work_buf_a, tbdest->name) ;
			strcat(work_buf_a, " := false") ; */
	
		}
	
		/*Let's make a recursive call with this*/
		rec_build_arcs(n, e, l->next, work_buf_a, work_buf_g) ;
		
		
		/* step 2 : the source variable is false */
		
		/*We first re-initialize the working copies of the guard and the assign*/
		strcpy(work_buf_a, buf_a) ;
		strcpy(work_buf_g, buf_g) ;	
		
		add_to_guard(work_buf_g, tbsource->name, 1) ;
		
/*		if (strlen(work_buf_g) != 0)
			strcat(work_buf_g, ", ") ;
		strcat(work_buf_g, "not(") ;
		strcat(work_buf_g, tbsource->name)  ;
		strcat(work_buf_g, ") ") ;*/
		
		/*Let's analyse the expression to see what to assign*/
		if (l->exp->not == 0) /*dest = source*/
		{
			add_to_action(work_buf_a, tbdest->name, _SFALSE_) ;
/*			if(strlen(work_buf_a) != 0)
				strcat(work_buf_a, ", ") ;
			strcat(work_buf_a, tbdest->name) ;
			strcat(work_buf_a, " := false") ; */ /*don't forget the source is false !*/
		}
		else /*dest = not(source)*/
		{
			add_to_action(work_buf_a, tbdest->name, _STRUE_) ;
/*			if(strlen(work_buf_a) != 0)
				strcat(work_buf_a, ", ") ;
			strcat(work_buf_a, tbdest->name) ;
			strcat(work_buf_a, " := true") ;*/ /*don't forget the source is false !*/
	
		}
	
		/*Let's make a recursive call with this*/
		rec_build_arcs(n, e, l->next, work_buf_a, work_buf_g) ;
		
	
	}
}

/* l1 contains the assignations involving only global variables */
/* list contains the rest of the assignation */
void subList(AssignRet ** list, AssignRet ** l1)
{
	AssignRet * next;
	AssignRet * ptr;
	*l1 = NULL;
	
	ptr = *list;
	*list = NULL;
	while (ptr != NULL) {
		next = ptr->next;
		if ((ptr->varType == _GLOBVAR) && (ptr->exp->type == _GLOBVAR)) {
			ptr->next = *l1;
			*l1 = ptr;
		}
		else {
			ptr->next = *list;
			*list = ptr;
		}		
		ptr = next;
	}
}

/* This function looks for the msg message in the message list l. If it is not present, it inserts it into the list */
void msgList_insert(char * msg, PTR_msgList * l) 
{
	PTR_msgList p = *l ;
	if (p == NULL) /* head insertion */
	{
		(*l) = (PTR_msgList )malloc(sizeof(msgList)) ;
		(*l)->next = NULL ;
		(*l)->message = (char *)malloc(sizeof(char)*(strlen(msg)+1)) ;
		strcpy((*l)->message, msg) ;
	}
	else
	{	
		while ((strcmp(p->message, msg)!= 0) && (p->next != NULL))
			p = p->next ;
		if (strcmp(p->message, msg) != 0) /* We haven't found the msg -> we tail-insert it */
		{
			p->next = (PTR_msgList )malloc(sizeof(msgList)) ;
			p->next->next = NULL ;
			p->next->message = (char *)malloc(sizeof(char)*(strlen(msg)+1)) ;
			strcpy(p->next->message, msg) ;
		}	
	}
}

